/**
 * @file njamd.h
 *
 * Main njamd header file. Contains ALL the data structures and defines needed
 * for NJAMD operation.
 *
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 */
#ifndef __NJAMD_LIB_MEM_H__
# define __NJAMD_LIB_MEM_H__
#include <config.h>
#define NJAMD_INFO PACKAGE VERSION " by Mike Perry\n"


/* Constants */
/** Number of bits in a megabyte */
#define SHIFT_MEGABYTE		20	
#ifndef TRUE
# define TRUE				1
# define FALSE				0
#endif

/**@{ @name Hardcoded config values */
/** Define this to the stack depth you want to trace */
#define TRACE_DEPTH     3

/** The limit on the core size when NJAMD_DUMP_CORE=soft is selected */
#define NJ_LIMIT_SOFTCORE		(4<<SHIFT_MEGABYTE)
/*@}*/

/* various debugging macros */
#ifdef EDEBUG
# define TRACEF 	write(2, __FUNCTION__"\n", strlen(__FUNCTION__)+1)
# define TR			do \
					{ \
						write(2, __FILE__":"__FUNCTION__"\n", strlen(__FILE__":"__FUNCTION__"\n")); \
						func = __FUNCTION__; line = __LINE__; file = __FILE__; \
					} while(0)
# define PRINTF(x...)	__nj_eprintf( __FUNCTION__ ": " ##x)
# define EPRINTF(x...)  __nj_eprintf( __FUNCTION__ ": " ##x)
# define __USE_UNIX98
#elif DEBUG
# define __USE_UNIX98
# define TRACEF 	write(2, __FUNCTION__"\n", strlen(__FUNCTION__)+1)
# define TR		do { file = __FILE__; func = __FUNCTION__; line = __LINE__; } while(0)
# define PRINTF(x...)	__nj_eprintf( __FUNCTION__ ": " ##x)
# define EPRINTF(x...)
#else
# define TRACEF
# define TR
# define PRINTF(x...)
# define EPRINTF(x...) 
#endif
	
#include <stdio.h>
#ifdef _THREAD_SAFE
# include <pthread.h>
#endif
#include <unistd.h>
#include <sys/param.h> /* stupid openbsd */
#include <limits.h>
#include <sys/types.h>
#include <signal.h>
#ifdef HAVE_SYS_USER_H
# include <sys/user.h>
#endif
#include <sys/mman.h>
#include <stdarg.h>


#define LINUX_PROC_OVERCOMMIT		"/proc/sys/vm/overcommit_memory"
#define LINUX_PROC_MAX_MAP_COUNT	"/proc/sys/vm/max_map_count"

/**@{ @name Methods of protecting freed memory */
#define ERROR_CHK_FREE		0
#define SEGV_CHK_FREE		1
#define NO_CHK_FREE			2
#define NO_FREE				3
/*@}*/

/**@{ @name Allocation methods */
#define ALLOC_OVER			0
#define ALLOC_UNDER			1
#define ALLOC_SUNDER		2
#define ALLOC_NONE			3
/*@}*/

/**@{ @name Dummy return addresses */ 
#define NJ_DESTRUCTOR			(-1)
#define NJ_RETADDR_NONE			(-3)
#define NJ_RETADDR_NOTFREE		(-5)
/*@}*/

/** The value to fencepost our buffers with to check for consistancy */
#define NJ_FENCEPOST			(0xdeadbeef)

	
/* Try to find NJ_PAGE_SIZE */
#if defined(HAVE_PAGE_SIZE)
# ifndef NJ_PAGE_SIZE
#  if defined(PAGE_SIZE)
#   define NJ_PAGE_SIZE PAGE_SIZE
#  elif defined(EXEC_PAGESIZE)
#   define NJ_PAGE_SIZE EXEC_PAGESIZE
#  elif defined(NBPG)
#   ifndef CLSIZE
#    define CLSIZE 1
#   endif
#   define NJ_PAGE_SIZE (NBPG*CLSIZE)
#  elif defined(NBPC)
#   define NJ_PAGE_SIZE (NBPC)
#  elif defined(PAGESIZE)
#   define NJ_PAGE_SIZE PAGESIZE
#  endif
# endif
#endif

#if defined(HAVE_PAGE_SHIFT)
# define NJ_PAGE_SHIFT PAGE_SHIFT
#endif

#if defined(HAVE_PAGE_MASK)
# ifdef NEGATE_PAGE_MASK
#  define NJ_PAGE_MASK			(~PAGE_MASK)
# else
#  define NJ_PAGE_MASK 			  PAGE_MASK
# endif
#endif


/* MAP ANONYMOUS */ 
#if !defined(MAP_ANONYMOUS) && defined(MAP_ANON)              
# define MAP_ANONYMOUS MAP_ANON                               
#endif                                                        
                                                              
#if defined(MAP_ANONYMOUS) && !defined(MAP_ANON)              
# define MAP_ANON MAP_ANONYMOUS                               
#endif          

#ifndef MAP_ANON
# define MAP_ANON 0
# define MAP_ANONYMOUS 0
#endif


/* Also in fe/njamd.h */
#ifndef HAVE_VSNPRINTF
# define vsnprintf(s, n, f, ap)			vsprintf(s, f, ap)
#endif

#ifndef HAVE_SNPRINTF
# define snprintf(s, n, f, ap...)		sprintf(s, f, ## ap)
#endif

/**@{ @name Alignment macros */
/** Align to a certain number of bytes, rounding up */
/** FIXME: If shit breaks on Solaris, look here first */
#define NJ_ALIGN_TO(val, alignto)				\
		((alignto > 1) ? (((u_long)val) + (alignto) - 1) & (~((alignto) - 1)) \
						: ((u_long)val))

/** Aligned objects should lack the lower bits */
#define NJ_IS_ALIGNED(val, alignto)		\
		!((u_long)(val) & ((alignto) - 1))

/** Overhead due to aligning */
#define NJ_ALIGN_OVERHEAD(val, alignto)	\
		(NJ_ALIGN_TO(val, alignto) - (u_long)(val))
/*@}*/
	
#define NJ_MIN(a,b)			((a) < (b) ? (a) : (b))
#define NJ_MAX(a,b)			((a) > (b) ? (a) : (b))

/** Exit without calling destructors of program, but clean up a bit */
#define nj_exit(ret)				\
	do { 							\
		if(NJ_ISSET(__nj_flags, NJ_PERSIST_HEAP))	\
		{ \
			*((struct mem_stats *)&__nj_heap_tbl.table[__nj_heap_tbl.n_entries]) = __nj_mem_stats; \
			msync((void *)__nj_heap_tbl.table, __nj_heap_tbl.n_entries*sizeof(struct heap_entry) + sizeof(struct mem_stats), \
					MS_SYNC); \
			truncate(__nj_heap_file, __nj_heap_tbl.n_entries*sizeof(struct heap_entry) + sizeof(struct mem_stats));\
		} \
		else if(!NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY)) \
			unlink(__nj_heap_file); \
		_exit(ret); \
	} while(0);

	
/* Data types */
/** Memory statistics data structure */
struct mem_stats
{
	u_int total_user;	 /**< Total user memory ever requested */
	u_int total_faulted; /**< Total memory faulted by system (pages present) */
	u_int peak_user;  	 /**< Peak user mem */
	u_int peak_faulted;  /**< Peak faulted mem */
	u_int peak_overhead; /**< Peak overhead due to NJAMD */
	u_int total_allocs;  /**< Total number of allocation requests */
	u_int leaked_allocs; /**< Total number of leaked allocs */
	u_long as_used;		 /**< Total amount of address space used */
	u_char alloc_type;   /**< Allocation type */
};

/** Description of a single slab table */
struct slab_desc
{
	void *addr;       /**< address of the slab */
	void *next;       /**< start of next chunk in the slab to slice out */
	size_t remainder; /**< Amount of space left in the slab */
};

/** The table of slabs */
struct slab_tbl
{
	int size;				/**< How many entries */
	struct slab_desc *tbl;	/**< The table itself */
};

/** 
 * A list of freed memory. Constructed from the freed segments themselves
 * when NO_CHK_FREE is set.
 */
struct free_list
{
	void *addr;				/**< The actual beginning of the segment */
	struct free_list *next; /**< The next pointer in the list */
};

/** Passed to the real pthread_create so we know all thread's starting
 * points.
 * @see pthread_create() __nj_pthread_launch() */
struct pthread_arg_wrapper
{
	void *(*start)(void *); /**< Always __nj_pthread_launch() */
	void *real_arg;			/**< The user's actual arg */
};

/** Alocation and deallocation calltraces for a block */
struct called 
{
	u_long alloced[TRACE_DEPTH];
	u_long freed[TRACE_DEPTH];
};

/** Heap entry. Describes both an open slot, and an occupied one */
struct heap_entry
{
	union addr_u {
		void *start; 			 /**< Location of buffer in memory */
		struct heap_entry *next; /**< next free slot (used to make a stack) */
	} u_addr;
	struct called loc; 			 /**< where we were alloc'ed / freed */
	size_t ulen; 				 /**< Length of the user segment */
};

/** Holds the heap table */
struct heap_tbl
{
	int n_entries;
	struct heap_entry *table;
};

/** The middle directory holds a table of tables */
struct heap_mid_dir
{
	int n_entries;
	struct heap_entry **table;
};

/** the top direcotry holds a table of middle tables */
struct heap_top_dir
{
	int n_entries;
	struct heap_entry ***table;
};

typedef void *(*nj_fcn_ptr_t)(void *, va_list);
typedef RETSIGTYPE (*nj_sighandler_t)(int);
#ifdef HAVE_POSIX_SIGACTION
typedef void (*nj_sigaction_t)(int, siginfo_t *, void *);
#endif
typedef void *(*nj_allocator_t)(void *, size_t, int, size_t, u_long *, char *);
typedef void (*nj_deallocator_t)(void *, u_long *);

#ifndef DOXGEN_PARSER
/* Prototypes */
/* mem_xx.c */
void *__nj_none_alloc		(void *, size_t, int, size_t, u_long *, char *);
void *__nj_overflow_alloc	(void *, size_t, int, size_t, u_long *, char *);
void *__nj_sunderflow_alloc	(void *, size_t, int, size_t, u_long *, char *);
void *__nj_underflow_alloc	(void *, size_t, int, size_t, u_long *, char *);
void __nj_none_free			(void *, u_long *);
void __nj_overflow_free		(void *, u_long *);
void __nj_sunderflow_free	(void *, u_long *);
void __nj_underflow_free	(void *, u_long *);

/* setup.c */
void __nj_primary_init		(void);
void __nj_secondary_init	(void);
void __nj_register_handlers	(void);

/* util.c */
int  __nj_dump_entry					(struct heap_entry *, int);
void __nj_free_list_push				(void *, void *, size_t);
struct heap_entry *__nj_new_heap_entry	(void *, int, u_long *);
void __nj_heap_stack_push				(struct heap_entry *);
int  __nj_is_valid_heap_ptr				(struct heap_entry *);
void __nj_eprintf						(const char *, ...);
void __nj_print_stack_trace				(u_long *, int);
void __nj_update_stats					(int, int, int);
void __nj_perror						(const char *);
void *__nj_free_list_pop				(size_t);
void *__nj_slab_fetch					(size_t);
int  __nj_address_from_malloc			(void *);
char *__nj_ptoa							(void *);
void *__nj_pthread_launch				(void *);
void __nj_ptr_info						(u_long);
void __nj_pthread_launch_end			(void);
struct heap_entry *__nj_heap_stack_pop	(void);
void __nj_userspace_ret					(u_long *, int);
int  __nj_entered_usercode				(void);
void __nj_prot_to_none					(void);
void __nj_out_of_memory					(void);
void __nj_usage_stats					(void);
struct heap_tbl __nj_new_heap_tbl		(void);
void __nj_dump_leaks					(int);
int  __nj_slab_alloc					(int);
char *__nj_itoa							(int);
	
/* public.c */
void __nj_pthread_exit_start			(void);
void __nj_pthread_exit_end				(void);
void __nj_exit_start					(void);
void __nj_exit_end						(void);

#endif /* DOXYGEN_PARSER */

/* Global variables */
#if defined(DEBUG) || defined(EDEBUG)
extern char 					*file;
extern char 					*func;
extern int 						line;
#endif

/**@{ @name global mm data structures */
extern struct heap_entry 		*__nj_heap_stack;
extern struct heap_tbl	 		__nj_heap_tbl;
extern struct heap_mid_dir		__nj_heap_mid_dir;
extern struct heap_top_dir		__nj_heap_top_dir;
#ifdef _THREAD_SAFE
extern pthread_mutex_t 			__nj_pthread_lock;
extern pthread_mutex_t 			__nj_heap_lock;
extern pthread_mutex_t 			__nj_slab_lock;
extern pthread_mutex_t 			__nj_list_lock;
#endif
extern struct slab_tbl	 		__nj_slab_tbl;
extern struct free_list			*__nj_free_tbl[4];
extern int 						__nj_zfd;
extern int 						__nj_efd;
#ifdef NJ_PAGE_SIZE
extern u_char 					__nj_zero_page[NJ_PAGE_SIZE];
#else
extern u_char					*__nj_zero_page;
#endif
extern int						__nj_anonfd;
/*@}*/

/**@{ @name Allocator functions */
extern nj_allocator_t			__nj_known_alloc;
extern nj_deallocator_t			__nj_known_free;
/*@}*/

/**@{ @name Prefrence variables */
extern int						__nj_display_depth;
extern int 						__nj_known_chk_free;
extern u_long 					__nj_align;
extern int						__nj_prot;
/*@}*/

/**@{ @name Status variables */
extern char		 				__nj_heap_file[80];
extern u_long					__nj_known_slab0;
extern u_long					__nj_slab0;
extern u_long					__nj_sbrk0;
extern struct mem_stats			__nj_mem_stats;
extern u_int 					__nj_leaked_user;
extern u_int 					__nj_leaked_faulted;
/*@}*/

/**@{ @name Flag macros */
#define NJ_SET(var, fl)			(var) |= (fl)
#define NJ_ISSET(var, fl)		((var) & (fl))
#define NJ_CLEAR(var, fl)		(var) &= ~(fl)
/*@}*/

/**@{ @name Prefrence flags.
 * @see __nj_secondary_init()
 */
#define NJ_NO_RETADDRS			(1<<0)
#define NJ_PERSIST_HEAP 		(1<<1)
#define NJ_STANDALONE			(1<<2)
#define NJ_TRACE_LIBS			(1<<3)
#define NJ_ALLOW_MALLOC0		(1<<4)
#define NJ_ALLOW_FREE0			(1<<5)
#define NJ_DUMP_STATS			(1<<6)
#define NJ_DUMP_HARDCORE		(1<<7)
#define NJ_DUMP_SOFTCORE		(1<<8)
#define NJ_NO_FREE_INFO			(1<<9)
/*@}*/

/**@{ @name Status flags */
#define NJ_RAN_PRIMARY 			(1<<26)
#define NJ_RAN_SECONDARY		(1<<27)
#define NJ_RAN_LIBC				(1<<28)
#define NJ_ENTERED_MAIN			(1<<29)
#define NJ_ACTIVE_PTHREADS		(1<<30)
#define NJ_ENTERED_DESTRUCT		(1<<31)
extern u_int					__nj_flags;
/*@}*/

/**@{ @name Signal handlers.
 * @see __nj_sig_dispatch() signal() sigaction() 
 */
extern nj_sighandler_t			__nj_sig_hdlr[sizeof(u_long)*8];
#ifdef HAVE_POSIX_SIGACTION
extern nj_sigaction_t			__nj_sig_actn[sizeof(u_long)*8];
#endif
/*@}*/

/* Portability */
#ifndef NJ_PAGE_SIZE
extern u_long 					NJ_PAGE_SIZE;
#endif

#ifndef NJ_PAGE_MASK
# ifdef NJ_PAGE_SIZE
#  define NJ_PAGE_MASK			(~(NJ_PAGE_SIZE-1))
# else
extern u_long 					NJ_PAGE_MASK;
# endif
#endif

#ifndef NJ_PAGE_SHIFT
extern u_long 					NJ_PAGE_SHIFT;
#endif

/* Macros that depend on data types or variables */
#define ADDRESS_SIZE_32		((u_long)0xffffffff)
#define NJ_ACCOUNTING_OVERHEAD		(sizeof(struct heap_size *))

/**@{ @name Heap table macros.
 * The heap table must accomodate enough heap_entries to describe the address
 * space. Worst case is if each heap_entry describes only 2 pages of memory 
 * This is approximately 16 megs on a 32bit address space. Don't worry, it's
 * only 16 megs of ADDRESS SPACE.. it's not used until faulted, and is synced
 * often, so not much of it is resident 
 *
 * Aha. Only create the supertable when the first 32 bit table is full.  This
 * is then independant of bit width, an allows us to do crazy shit like store
 * all info even when it's being freed (SAVE_ALL_INFO). This should be done as
 * 32, 16, and 16, with the last 16 not being used on Linux at least.
 * @see __nj_new_heap_tbl() save_heap()
 */
#define HEAP_TBL_NENTRIES			(ADDRESS_SIZE_32 >> (NJ_PAGE_SHIFT+1))
#define HEAP_TBL_SIZE				(HEAP_TBL_NENTRIES \
										* sizeof(struct heap_entry) \
										+ sizeof(struct mem_stats))
/** Each level of metatables handles 16 bits of additional address space.
 * 2^32 * 2^16 * 2^16 = 2^64 */
#define HEAP_META_TBL_NENTRIES		(1<<16)
#define HEAP_META_TBL_SIZE			(HEAP_META_TBL_NENTRIES*sizeof(void *))

	/** 
 * Determines if an entry is free. We store a special return address
 * (NJ_RETADDR_NOTFREE) in the freed stacktrace if an entry is not free.
 * @see __nj_new_heap_entry()
 */
#define entry_is_freed(entry) \
	((entry)->loc.freed[1] != NJ_RETADDR_NOTFREE)
/*@}*/

/**@{ @name Size calculation macros */	
/** Round len up to nearest page boundry */
#define ROUND_TO_PAGE(len)			((((len)-1) & NJ_PAGE_MASK) + NJ_PAGE_SIZE)

/** Round len with accounting info to a blocksize usable in NJAMD */
#define ROUND_TO_BLOCK(len)			((((len)-1) & NJ_PAGE_MASK) + \
															(2<<NJ_PAGE_SHIFT))

/** Round len with accounting info to a page boundary */
#define ROUND_TO_FAULT(len)			((((len)+NJ_ACCOUNTING_OVERHEAD-1) \
										& NJ_PAGE_MASK) + NJ_PAGE_SIZE)
/*@}*/

/**@{ @name Slab macros 
 * The slab does not need a metatable because it can be safely and efficiently
 * expanded (nothing points into it)
 * @see __nj_slab_fetch() __nj_slab_resize() __nj_slab_alloc()
 */
/** 
 * Size of a slab. 32 megs for 32 bit processor, 128 megs for 64
 * @FIXME: Change this as RAM gets cheaper/more abundant (it really should be
 * an exponential fit, like 32 and 512) 
 */
#define SLAB_MEGS			(sizeof(void *)*24 - 64)
#define SLAB_SIZE			(SLAB_MEGS << SHIFT_MEGABYTE)

#define SLAB_TABLE_NENTRIES		ROUND_TO_PAGE(ADDRESS_SIZE_32 / SLAB_SIZE + 1)
#define SLAB_TABLE_SIZE			(SLAB_TABLE_NENTRIES*sizeof(struct slab_desc))
/*@}*/

/**@{ @name Preinit settings
 * Settings to use before we can read anything from the environment.  Don't
 * set these to __nj_none_alloc() and free, because we need these allocations
 * to take place in the slabs so we can determine if the memory is from the
 * preinit allocator when we free memory later.
 * @see realloc() free() 
 */
#define PREINIT_ALLOCATOR		__nj_overflow_alloc
#define PREINIT_DEALLOCATOR		__nj_overflow_free
/** 
 * Don't change this to NO_CHK_FREE.
 * As tempting as it is, if you set this to NO_CHK_FREE, shit will break,
 * because we will have preinit memory in the free lists, among other things. 
 * @see nj_free_fini()
 */
#define PREINIT_CHK_FREE		SEGV_CHK_FREE
/*@}*/

/**@{ @name Default settings. 
 * These can be anything. */	
#define DEFAULT_ALLOCATOR		__nj_overflow_alloc
#define DEFAULT_DEALLOCATOR		__nj_overflow_free
#define DEFAULT_CHK_FREE		SEGV_CHK_FREE
#define DEFAULT_ALLOC_TYPE		ALLOC_OVER /** Must match ALLOCATOR */
/*@}*/


#endif /* __NJAMD_LIB_MEM_H__ */

// vim:ts=4
