/** 
 * @file libc.c
 * 
 * Resololves symbols we need from libc.
 * 
 * Even when not using NJAMD_PROT=none, NJAMD still needs some symbols from
 * libc to be able to handle pthreads, and to properly install signal
 * handlers in a manner transparent to the user.
 * 
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 */
#include <lib/njamd.h>
#include <lib/libc.h>
#ifdef HAVE_SANE_DLOPEN
# include <dlfcn.h>
#endif
#ifdef HAVE_DL_ITERATE_PHDR
# include <link.h>
# include <stddef.h>
# include <string.h>
#endif
#include <stdio.h>
#include <errno.h>

nj_sighandler_t (*__njLibcSignal)(int, nj_sighandler_t);
#ifdef HAVE_POSIX_SIGACTION
int (*__njLibcSigaction)(int, const struct sigaction *, struct sigaction *);
#endif

#ifdef _THREAD_SAFE
int  (*__nj_pthread_create)(pthread_t  *  thread, pthread_attr_t *attr, 
		void * (*start_routine)(void *), void *);
void (*__nj_pthread_exit)(void *);
#endif
void (*__nj_exit)(int);

void (*__nj_libc_free)(void *);
void *(*__nj_libc_malloc)(size_t);
void *(*__nj_libc_calloc)(size_t, size_t);
void *(*__nj_libc_realloc)(void *,size_t);

#ifdef HAVE_DL_ITERATE_PHDR
static const char *libc_so_path, *libpthread_so_path;

static int
__njLibcIterateCallback (struct dl_phdr_info * info, size_t size, void * ptr)
{
	const char *p;
	if (size < offsetof(struct dl_phdr_info, dlpi_phdr))
		return -1;
	if (info->dlpi_name == NULL)
		return 0;
	p = strstr(info->dlpi_name, "/libc.so.");
	if (p != NULL)
	{
		if (! libc_so_path
		    || strncmp (info->dlpi_name, "/lib/", 5) == 0)
			libc_so_path = info->dlpi_name;
	}
	else
	{
		p = strstr (info->dlpi_name, "/libpthread.so.");
		if (p != NULL)
		{
			if (! libpthread_so_path
			    || strncmp (info->dlpi_name, "/lib/", 5) == 0)
				libpthread_so_path = info->dlpi_name;
		}
	}
	return 0;
}
#endif

static int
__nj_dummy_pthread_create(pthread_t * thread, pthread_attr_t *attr,
			  void * (*start_routine)(void *), void * arg)
{
	__nj_eprintf("NJAMD: libpthread was not loaded at program start\n");
	return ENOSYS;
}

static void
__nj_dummy_pthread_exit(void *retval)
{
	__nj_eprintf("NJAMD: libpthread was not loaded at program start\n");
}

/** 
 * Handles initialization of dynamically loaded functions.
 *
 * This function uses dlopen to resolve all the functions we need if we plan
 * to use NJAMD_PROT=none, as well as to keep track of the threads and when
 * they exit (by wrapping exit and pthread_exit)
 *
 * @returns The error string reported by dlerror
 * @see pthread_exit() exit() signal() sigaction()
 */ 
const char *__njLibcFuncInit(void)
{
#ifdef HAVE_SANE_DLOPEN
	void *libc_handle;
# ifdef _THREAD_SAFE
	void *pthread_handle;
# endif
	const char* error = NULL;
#endif
	if (NJ_ISSET(__nj_flags, NJ_RAN_LIBC))
		return NULL;
	TR;

#ifdef HAVE_DL_ITERATE_PHDR
	if (dl_iterate_phdr (__njLibcIterateCallback, NULL) < 0)
		return "Couldn't iterate over shared libraries";
	if (libc_so_path == NULL)
		return "Couldn't find libc.so";
#undef LIBC_SO_PATH
#undef LIBPTHREAD_PATH
#define LIBC_SO_PATH libc_so_path
#define LIBPTHREAD_PATH libpthread_so_path
#endif

#ifdef HAVE_SANE_DLOPEN
	
#ifdef _THREAD_SAFE
	if (LIBPTHREAD_PATH)
		pthread_handle = dlopen(LIBPTHREAD_PATH, RTLD_LAZY);
	else
	{
		pthread_handle = NULL;
		__nj_pthread_create = __nj_dummy_pthread_create;
		__nj_pthread_exit = __nj_dummy_pthread_exit;
	}
#endif

	TR;
	/* If we have a libpthread, then standard libc functions are threadsafe.
	   If not, then we get the libc functions from the threadsafe lib if we're
	   building threadsafe version, or the standard path if not. Currently we
	   never define _THREAD_SAFE */
#ifdef HAVE_LIBPTHREAD
	if((libc_handle = dlopen(LIBC_SO_PATH, RTLD_LAZY)) == NULL)
	{
		__nj_eprintf("libc.so not found\n");
		return dlerror();
	}
#elif defined(_THREAD_SAFE) 
	if((libc_handle = dlopen(LIBC_R_PATH, RTLD_LAZY)) == NULL)
	{
		__nj_eprintf("libc.so not found\n");
		return dlerror();
	}
#else
	if((libc_handle = dlopen(LIBC_SO_PATH, RTLD_LAZY)) == NULL)
	{
		__nj_eprintf("libc.so not found\n");
		return dlerror();
	}
#endif

	TR;
	if((__njLibcSignal = dlsym(libc_handle, "signal")) == NULL)
	{
		error = dlerror();
		goto done;
	}
#ifdef HAVE_POSIX_SIGACTION
	if((__njLibcSigaction = dlsym(libc_handle, "sigaction")) == NULL)
	{
		error = dlerror();
		goto done;
	}
#endif

	if((__nj_exit = dlsym(libc_handle, "exit")) == NULL)
	{
		error = dlerror();
		goto done;
	}

	/* We have to resolve all of the malloc stuff because we don't find out if
	 * we need them until later */
	if((__nj_libc_malloc = dlsym(libc_handle, "malloc")) == NULL)
	{
		error = dlerror();
		goto done;
	}

	if((__nj_libc_calloc = dlsym(libc_handle, "calloc")) == NULL)
	{
		error = dlerror();
		goto done;
	}
	
	if((__nj_libc_realloc = dlsym(libc_handle, "realloc")) == NULL)
	{
		error = dlerror();
		goto done;
	}
	
	if((__nj_libc_free = dlsym(libc_handle, "free")) == NULL)
	{
		error = dlerror();
		goto done;
	}

#ifdef _THREAD_SAFE
	if(pthread_handle)
	{
		if((__nj_pthread_create = dlsym(pthread_handle, "pthread_create")) == NULL)
		{
			error = dlerror();
			goto done;
		}

		if((__nj_pthread_exit = dlsym(pthread_handle, "pthread_exit")) == NULL)
		{
			error = dlerror();
			goto done;
		}
	}
#endif

	NJ_SET(__nj_flags, NJ_RAN_LIBC);
	
	TR;
done:
	dlclose(libc_handle);
#ifdef _THREAD_SAFE
	if(pthread_handle)
		dlclose(pthread_handle);
#endif
	return error;
#endif
	__njLibcSignal = signal;
	return NULL;
}
// vim:ts=4
