/** 
 * @file mem_none.c
 *
 * Provides the malloc wrapping capabilities of NJAMD.
 * 
 * When NJAMD_PROT=none, these functions are used to allocate and free memory.
 * No protection is done, but fencepost values are set, and memory accounting
 * is performed.
 * 
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 */
#include <lib/njamd.h>
#include <lib/libc.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <errno.h>

/**
 * General purpose allocator.
 *
 * Gets memory from libc malloc or calloc, sets up the NJAMD accounting
 * buffers, and returns it.
 * 
 * @param ptr A previously allocated pointer for realloc
 * @param len The length to allocate
 * @param zero_fill If true, calloc
 * @param align Alignment (not yet used)
 * @param callstack The return addresses of the malloc call
 * @param caller The caller name, for error reporting.
 *
 * @sideeffects The heap table is modified (thru __nj_new_heap_entry()) 
 * 				Updates NJAMD statistics thru __nj_update_stats().
 * @returns A libc m/calloc'ed segment.
 * 
 * @FIXME: Alignment is much easier than I thought.. simply put the inf
 * pointer 1 NJ_PTR_aligned segment below (ie, and off the align mask) the return
 * value, which is aligned to anything. Then, the address we pass to free is
 * inf->start.
 */
void *__nj_none_alloc(void *ptr, size_t len, int zero_fill, size_t align,
		u_long *callstack, char *caller)
{
	struct heap_entry *inf;
	void *buf;
	void *ret;
	

	if(!len && NJ_ISSET(__nj_flags, NJ_ALLOW_MALLOC0))
	{
		return NULL;
	}

	/* Ask a stupid question, get a stupid answer */
	if(len <= 0)
	{
		__nj_eprintf("NJAMD/%s: Request must be > 0 (%d)\n", caller, len);
		__nj_print_stack_trace(callstack, TRACE_DEPTH);
		errno = EINVAL;
		return NULL;
	}


	if(!__nj_libc_realloc || !__nj_libc_malloc || !__nj_libc_calloc)
	{
		__nj_eprintf("NJAMD_PROT=none can't be used on this system (no "
			" dlopen?)\n");
		exit(1);
	}

	TR;
	/* Realloc? */
	if(ptr != NULL)
	{
		inf = *(struct heap_entry **)NJ_ALIGN_TO(ptr-NJ_ACCOUNTING_OVERHEAD, NJ_PTR_ALIGNMENT);
		TR;

		if(!__nj_is_valid_heap_ptr(inf))
		{
			__nj_eprintf("NJAMD/%s: Heap corruption detected on %s of 0x%lx.\n\tTry using strict underflow option\n",
					caller, caller, (u_long)ptr);
			raise(SIGSEGV);
			exit(1);
		}

		TR;

		if(*(u_int *)(ptr + NJ_ALIGN_TO(inf->ulen, NJ_INT_ALIGNMENT)) != NJ_FENCEPOST)
		{
			__nj_eprintf("NJAMD/%s: Heap corruption. Try using overflow option\n", caller);
			memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));
			__nj_dump_entry(inf, TRACE_DEPTH);
			raise(SIGSEGV);
			exit(1);
		}
		/* Report the old memory as freed */	
		__nj_update_stats(-inf->ulen, 
			-(((u_long)ptr - (u_long)inf->u_addr.start) + NJ_ALIGN_TO(inf->ulen, NJ_INT_ALIGNMENT) + sizeof(int)), 
			-(((u_long)ptr - (u_long)inf->u_addr.start) + NJ_ALIGN_TO(inf->ulen, NJ_INT_ALIGNMENT) + sizeof(int))); 
		
		/* Ohh fuck.. the alignment is gonna mean the users shit starts somewhere else */
		buf = __nj_libc_realloc(inf->u_addr.start, NJ_ALIGN_TO(len, NJ_INT_ALIGNMENT) + (align - 1) + NJ_ACCOUNTING_OVERHEAD + sizeof(int));

		ret = (void *)NJ_ALIGN_TO(buf, align);
		if(((u_long)ret - (u_long)buf) < NJ_ACCOUNTING_OVERHEAD)
			ret = (void *)NJ_ALIGN_TO(buf + NJ_ACCOUNTING_OVERHEAD, align);

		/* If the alignment has changed, we must realloc */
		if(align != (u_long)ptr - (u_long)inf->u_addr.start)
		{
			memmove(ret, buf + (u_long)ptr - (u_long)inf->u_addr.start, NJ_MIN(len, inf->ulen));
		}
		
		/* The following is handled in __nj_new_heap_entry(), but we don't
		 * want two heap entries describing what might be the same region of
		 * memory due to the realloc.. */
		inf->ulen = len;
		memcpy(inf->loc.alloced, callstack, TRACE_DEPTH*sizeof(u_long));
		inf->u_addr.start = buf;
	}
	else
	{
		TR;
		/* We have to guess pessimisticly at these alignment overheads because we don't know what malloc will return */
		if(zero_fill)
			buf = __nj_libc_calloc(1, NJ_ALIGN_TO(len, NJ_INT_ALIGNMENT) + (align - 1) + NJ_ACCOUNTING_OVERHEAD + sizeof(int));
		else
			buf = __nj_libc_malloc(NJ_ALIGN_TO(len, NJ_INT_ALIGNMENT) + (align - 1) + NJ_ACCOUNTING_OVERHEAD  + sizeof(int));
		TR;
		inf = __nj_new_heap_entry(buf, len, callstack);
	
		ret = (void *)NJ_ALIGN_TO(buf, align);
		if(((u_long)ret - (u_long)buf) < NJ_ACCOUNTING_OVERHEAD)
			ret = (void *)NJ_ALIGN_TO(buf + NJ_ACCOUNTING_OVERHEAD, align);
	
	}
	TR;

	/* lost space at this point <= NJ_ACCOUNTING_OVERHEAD + align - 1 */


#ifdef DEBUG
	/* This line will only execute if ret is aligned to < NJ_PTR_ALIGNMENT,
	 * or unless NJ_ACOUNTING_OVERHEAD < NJ_PTR_ALIGNMENT.
	 * Won't ever happen because malloc wouldn't return an unaligned buf,
	 * and since no arch in the world aligns pointers to larger than their size.
	 */
	if(NJ_ALIGN_TO(ret-NJ_ACCOUNTING_OVERHEAD, NJ_PTR_ALIGNMENT)+NJ_ACCOUNTING_OVERHEAD 
			- (u_long)buf < NJ_ACCOUNTING_OVERHEAD)
		__nj_eprintf("Warning.. Email mikepery@fscked.org\n");
#endif
	
	/* put the inf pointer at the first ptr aligned chunk BEFORE ret */
	*(struct heap_entry **)NJ_ALIGN_TO(ret-NJ_ACCOUNTING_OVERHEAD, 
			NJ_PTR_ALIGNMENT) = inf;

	__nj_update_stats(len, (align - 1) + NJ_ACCOUNTING_OVERHEAD + NJ_ALIGN_TO(len, NJ_INT_ALIGNMENT) + sizeof(int), 
			(align - 1) + NJ_ACCOUNTING_OVERHEAD + NJ_ALIGN_TO(len, NJ_INT_ALIGNMENT) + sizeof(int)); 
	
	TR;
	/* Used to check for overflow at free. The reason why we don't have a
	 * fencepost at the beginning is because we have the pointer there, and
	 * __nj_is_valid_heap_ptr will tell us if it was damaged */
	/** @FIXME: this shit will break stuff maybe... it might overflow.*/
	*(u_int *)(ret + NJ_ALIGN_TO(len, NJ_INT_ALIGNMENT)) = NJ_FENCEPOST;

	return ret;
}

/**
 * Returns memory to the libc malloc heap.
 *
 * Also checks if the fencepost values have been violated.
 * 
 * @param buf Allocated buffer to free
 * @param callstack Return address call stack
 *
 * @sideeffects Statistics are updated thru __nj_update_stats(), and the
 * 				segment info pointer may be reused with __nj_heap_stack_push().
 */ 
void __nj_none_free(void *buf, u_long *callstack)
{
	struct heap_entry *inf;

	if(buf == NULL || __nj_known_chk_free == NO_FREE)
	{
		return;
	}

	TR;
	
	/* see above */
	inf = *(struct heap_entry **)NJ_ALIGN_TO(buf-NJ_ACCOUNTING_OVERHEAD, NJ_PTR_ALIGNMENT);

	TR;
	
	if(!__nj_is_valid_heap_ptr(inf))
	{
		__nj_eprintf("NJAMD/free: Heap corruption detected on free of 0x%lx.\n\tTry using strict underflow option\n",
				(u_long)buf);
		raise(SIGSEGV);
		exit(1);
	}

	TR;
	
	if(*(u_int *)(buf + NJ_ALIGN_TO(inf->ulen, NJ_INT_ALIGNMENT)) != NJ_FENCEPOST)
	{
		__nj_eprintf("NJAMD/free: Heap corruption. Try using overflow option\n");
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));
		__nj_dump_entry(inf, TRACE_DEPTH);
		raise(SIGSEGV);
		exit(1);
	}
	
	/* If the user doesn't want information about freed segments, don't bother
	 * to copy */
	if(!NJ_ISSET(__nj_flags, NJ_NO_FREE_INFO))
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));
	
	__nj_update_stats(-inf->ulen, 
			-(((u_long)buf - (u_long)inf->u_addr.start) + NJ_ALIGN_TO(inf->ulen, NJ_INT_ALIGNMENT) + sizeof(int)), 
			-(((u_long)buf - (u_long)inf->u_addr.start) + NJ_ALIGN_TO(inf->ulen, NJ_INT_ALIGNMENT) + sizeof(int))); 
	
	TR;
	/* If we don't keep freed info, return it to the heap-stack */
	if(NJ_ISSET(__nj_flags, NJ_NO_FREE_INFO))
		__nj_heap_stack_push(inf);
	
	TR;
	if(__nj_libc_free)
		__nj_libc_free(inf->u_addr.start);
	else
	{
		__nj_eprintf("NJAMD_PROT=none can't be used on this system (no "
			" dlopen?)\n");
		exit(1);
	}
}

// vim:ts=4
