/** 
 * @file mem_over.c 
 *
 * Provides the overflow capabilities of NJAMD
 *
 * These functions all handle the case where a protect page of memory is
 * placed AFTER the user's buffer to prevent overflow. Fencepost values are
 * also placed before the buffer, and are checked upon free. These are the
 * default allocators.
 *
 * @NOTES: The memory layout for this is as follows: Info pointer on a page
 * boundary, an empy space, then the user's aligned memory starts, then the
 * protected page.
 * 
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 */
#include <lib/njamd.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <errno.h>

/**
 * Performs the first half of free/realloc. This includes retrieving an info
 * pointer from the segment, as well as checking fencepost values.
 *
 * @param start The first pagesize boundary below userbuf
 * @param userbuf The address the user tried to free
 * @param callstack The return address call stack
 * @param caller The ASCII string of the calling function (ie "free",
 *               "realloc") for error printing.
 * 
 * @sideeffects In some cases the permissions on start change to readonly.
 * @returns The heap entry info pointer corresponding to this memory segment.
 */
static struct heap_entry *nj_free_init(void *start, void *userbuf, 
		void *callstack, char *caller)
{
	struct heap_entry *inf = NULL;
	
	TR;
	/* We don't check for page aligned memory because it's guaranteed */

	
	/* Check to see if we had to tack on another page in postmalloc. If the
	 * userbuf and the start differ by less than the amount needed to do
	 * accounting, we then had to increase the allocated segment size by 1
	 * page. */
	if(userbuf - start < NJ_ACCOUNTING_OVERHEAD + sizeof(u_int))
	{
		start -= NJ_PAGE_SIZE;
	}

	/** 
	 * @FIXME: this may cause problems when known free differes from default,
	 * and system setup memory is actually freed. So far this hasn't presented
	 * itself. In that case, we will get false reports of heap corruption. 
	 * */
	if(__nj_known_chk_free == ERROR_CHK_FREE)
	{
		/* We have to mprotect this in case this WAS a double free, to avoid
		 * segfault */
		if(mprotect(start, NJ_PAGE_SIZE, PROT_READ) 
				== -1)
		{
			TR;
			__nj_perror(__FUNCTION__"/mprotect");
			raise(SIGSEGV);
			return NULL;
		}
	
		TR;

		/* A NULL inf can only mean read from freed memory */
		if(*(struct heap_entry **)start == 0)
		{
			/* However, if the rest of the page isn't 0, we might have an
			 * underflow of 0 */
			if(memcmp(start, __nj_zero_page, NJ_PAGE_SIZE) != 0)
			{
				__nj_eprintf("NJAMD: heap corruption. Try using the underflow "
						"option to pinpoint source of error\n");	
				raise(SIGSEGV);
				return NULL;
			}

			__nj_eprintf("NJAMD/%s: Double %s of address 0x%lx\n", caller, caller, (u_long)userbuf);
			raise(SIGSEGV);
			return NULL;
		}
	}

	TR;

	/* Index into the heap table */
	inf = *(struct heap_entry **)start;

	TR;

	if(!__nj_is_valid_heap_ptr(inf))
	{
		__nj_eprintf("NJAMD/%s: Heap corruption detected on %s of 0x%lx.\n\tTry using strict underflow option\n",
				caller, caller, (u_long)userbuf);
		raise(SIGSEGV);
		exit(1);
	}

	/* There's always room for fenceposts */
	if(*(u_int *)(start+NJ_ACCOUNTING_OVERHEAD) != NJ_FENCEPOST)
	{
		__nj_eprintf("NJAMD/%s: heap corruption. Try using the underflow "
						"option to pinpoint source of error\n", caller);	
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));
		__nj_dump_entry(inf, TRACE_DEPTH);
		raise(SIGSEGV);
		return NULL;
	}


	if(inf->u_addr.start != start)
	{
		__nj_eprintf("NJAMD/%s: Inconsistancy. Either address passed (0x%lx) "
				"is not directly from malloc, or small underflow. Try using "
				"strict underflow option. \n", caller, userbuf);
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));
		__nj_dump_entry(inf, TRACE_DEPTH);
		raise(SIGSEGV);
		return NULL;
	}

	if(!NJ_ISSET(__nj_flags, NJ_NO_FREE_INFO))
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));

	TR;
	return inf;
}

/** 
 * Performs the second half of free/realloc. Remaps memory pointed to
 * by start as specified by the NJAMD_CHK_FREE setting.
 *
 * @param start The actual start of the mmaped region
 * @param offset the length of the offset into the user's region.
 * @param inf The heap info pointer for this region
 *
 * @sideeffects Stats are updated. Start becommes remmaped or assigned to the
 * 				free list.
 */
static void nj_free_fini(void *start, size_t offset, struct heap_entry *inf)
{
	int fetched = ROUND_TO_BLOCK(inf->ulen);
	int chk_free = __nj_known_chk_free;
	int align_offset;

	/* 
	 * Since this fetched is wrong (it's based on an unaligned length) we can
	 * get the aligned offset by using substraction. 
	 * offset = real_fetch - aligned_len - NJ_PAGE_SIZE.
	 */
	/** @FIXME: don't actually need an align_offset variable */
	align_offset = fetched - NJ_PAGE_SIZE - offset - inf->ulen;
	if(align_offset < 0)
	{
#ifdef DEBUG
		__nj_eprintf("Good, align_offset is %d for a len of %d. Adjusting to %d. Offset: %d\n",
				align_offset, inf->ulen, align_offset + NJ_PAGE_SIZE, offset);
#endif
		fetched += NJ_PAGE_SIZE;
	}

	/* Check if allocated before we knew the protection scheme. 
	 * slab0 is the address of the first slab, and known_slab0 is the address
	 * where we first learned the allocation strategy */
	if((__nj_slab0 <= (u_long)start && (u_long)start < __nj_known_slab0))
		chk_free = PREINIT_CHK_FREE;

	/* We pass in negative arguments to subtract that lengh. But we only
	 * subtract the address space if NO_CHK_FREE is set. */
	__nj_update_stats(-inf->ulen, -ROUND_TO_FAULT(inf->ulen), (chk_free == NO_CHK_FREE) ? -ROUND_TO_BLOCK(inf->ulen) : 0);

	/* If we don't save info about freed segments, return them to the
	 * heap-stack */
	if(NJ_ISSET(__nj_flags, NJ_NO_FREE_INFO))
		__nj_heap_stack_push(inf);
	
	TR;
	switch(chk_free)
	{
		/* To check for double frees, we map a page of zero mem to clue us in */
		case ERROR_CHK_FREE:
			if(mmap(start, NJ_PAGE_SIZE, __nj_prot, MAP_FIXED | MAP_SHARED,
						__nj_zfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			if(mmap(start+NJ_PAGE_SIZE, fetched-NJ_PAGE_SIZE, __nj_prot, 
						MAP_FIXED | MAP_PRIVATE | MAP_ANON, __nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		case SEGV_CHK_FREE:
			if(mmap(start, fetched, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		case NO_CHK_FREE:
			TR;
			/* Q: what if the allocation strategy is unknown? 
			 * We might end up returning a write protected segment to the
			 * wrong allocator */
			/* A: Well, we don't find out that we should use NO_CHK_FREE until
			 * AFTER we know the allocation stratagy.. So we will ONLY free
			 * segments from the known allocator type in this way. Likewise,
			 * the unknown allocator type will never be used after this
			 * (because PREINIT_CHK_FREE is forbidden to be NO_CHK_FREE) */
			__nj_free_list_push(start, start, fetched);	
			
			break;

		case NO_FREE:
			/* Do nothing */
			break;

		default:
			__nj_eprintf("Unknown free checking option %d\n", chk_free);
			if(mmap(start, fetched, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;
	}


	TR;
}

/**
 * Provides allocated segments bounded by a protected page after. This
 * function handles malloc, calloc, and realloc. At least two pages are 
 * allocated by this function, but care is taken so that only one is faulted.
 * Thus, while the address space loss per alloc is 8k, the memory loss is less
 * than 4k.  (Unfortunately averaging around 3.9k) 
 *
 * @param ptr A previously allocated pointer for realloc
 * @param len The length to allocate
 * @param zero_fill If true, calloc
 * @param align Alignment of the returned buffer.
 * @param callstack The return addresses of the malloc call
 * @param caller The caller name, for error reporting.
 *
 * @sideeffects Free list is modified if NO_CHK_FREE. The heap table is modified 
 * 				(new info ptr). Updates NJAMD statistics.
 * @returns A mmaped, aligned, and protected segment.
 */
void *__nj_overflow_alloc(void *ptr, size_t len, int zero_fill, size_t align, 
		u_long *callstack, char *caller)
{
	struct heap_entry *inf;
	size_t fetch, offset;
	int protect_mem = FALSE;
	size_t aligned_len;
	void *buf = NULL;

	if(!len && NJ_ISSET(__nj_flags, NJ_ALLOW_MALLOC0))
	{
		return NULL;
	}

	/* Ask a stupid question, get a stupid answer */
	if(len <= 0)
	{
		__nj_eprintf("NJAMD/%s: Request must be > 0 (%d)\n", caller, len);
		__nj_print_stack_trace(callstack, TRACE_DEPTH);
		errno = EINVAL;
		return NULL;
	}

	/* By making len aligned by the specified alignment, the resulting buffer
	 * will be aligned */
	aligned_len = NJ_ALIGN_TO(len, align);

	/* We need enough mem to fit an aligned lenth, the inf pointer, and the 
	 * fencepost before the buffer*/
	fetch = ROUND_TO_BLOCK(aligned_len+NJ_ACCOUNTING_OVERHEAD+sizeof(u_int));
	/* The offset into our region where the user's memory starts. Since len 
	 * is aligned, and fetch and PAGE_SIZE are aligned, the start will be too */
	offset = fetch - aligned_len - NJ_PAGE_SIZE;
	
	TR;
	/* Realloc? */
	if(ptr != NULL)
	{
		void *start = (void *)((long)ptr & NJ_PAGE_MASK);
	
		TR;	
		if(!(inf = nj_free_init(start, ptr, callstack, caller)))
		{
			/* This should not happen */
			return NULL;
		}

		/* the actual start of the region is where the inf says */
		start = inf->u_addr.start;
		TR;
		
		/* Get a chunk from the slab allocator */
		if(__nj_known_chk_free != NO_CHK_FREE || 
				(buf = __nj_free_list_pop(fetch)) == NULL)
		{
			buf = __nj_slab_fetch(fetch);
			/* protect mem that comes from the slab */
			protect_mem = TRUE;
		}
		/* copy the buffer and appropriate accounting info */
		memcpy(buf + offset, ptr, NJ_MIN(inf->ulen, len));

		TR;
		/* Remap the old memory to bupkes */
		nj_free_fini(start, (u_long)ptr - (u_long)start, inf);

		*(struct heap_entry **)buf = inf = __nj_new_heap_entry(buf, len, callstack);
		start = buf;
		TR;

	}
	else
	{
		TR;
		/* Get a chunk from the slab allocator */
		if(__nj_known_chk_free != NO_CHK_FREE || 
				(buf = __nj_free_list_pop(fetch)) == NULL)
		{
			buf = __nj_slab_fetch(fetch);
			/* protect mem that comes from the slab */
			protect_mem = TRUE;
		}

		/* calloc */
		if(zero_fill)
		{
			int i;
			/* Page copies may be optimized by the lib */
			for(i=0; i < (fetch >> NJ_PAGE_SHIFT) - 1; i++)
				memcpy(buf + (i<<NJ_PAGE_SHIFT), __nj_zero_page, NJ_PAGE_SIZE); 
		}
		TR;
		*(struct heap_entry **)buf = inf = __nj_new_heap_entry(buf, len, callstack);
	}

	/* report the new memory */
	__nj_update_stats(len, fetch-NJ_PAGE_SIZE, fetch);

	if(protect_mem)
		if(mprotect(buf+fetch-NJ_PAGE_SIZE, NJ_PAGE_SIZE, __nj_prot) == -1)
		{
			__nj_perror(__FUNCTION__"/mprotect");
			exit(1);
		}

	TR;

	/* To check for underflows */
	/* There is always room for this now, we make sure to fetch enough for it */
	*(u_int *)(buf+NJ_ACCOUNTING_OVERHEAD) = NJ_FENCEPOST;

	return buf+offset;
}

/** 
 * Main entry point for freeing memory. Basically just calls __nj_free_init
 * and __nj_free_fini. 
 *
 * @param buf Allocated buffer to free
 * @param ret Return address call stack
 */ 
void __nj_overflow_free(void *buf, u_long *callstack)
{
	void *start = (void *)((u_long)buf & NJ_PAGE_MASK);
	struct heap_entry *inf;

	if(buf == NULL)
	{
		return;
	}

	TR;	
	if(!(inf = nj_free_init(start, buf, callstack, "free")))
	{
		/* Shouldn't happen */
		return;
	}
	
	
	/* the actual start of the region is where inf says */
	start = inf->u_addr.start;

	nj_free_fini(start, (u_long)buf - (u_long)start, inf);
}
// vim:ts=4
