/** 
 * @file mem_sunder.c 
 *
 * Provides the strict underflow capabilities of NJAMD
 *
 * These functions handle the case where a protected page of memory is placed
 * DIRECTLY before the user's allocated segment. This 0-byte boundary on
 * overflow comes at a price, namely an extra faulted page of memory, as well
 * as speed. This protection scheme is only advised if the standard underflow
 * scheme reported heap corruption yet no segmentation fault.
 *
 * @NOTES: This allocation scheme has memory like this: Protected page,
 * unprotect page. At the beginning of the protected page is the inf pointer.
 * The user's buf starts right at the second page, so needn't be aligned.
 * 
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 */
#include <lib/njamd.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <errno.h>


/**
 * Performs the first half of free/realloc. This includes retrieving an info
 * pointer from the segment, as well as checking fencepost values.
 *
 * @param start The first pagesize boundary below userbuf
 * @param callstack The return addresses callstack
 * @param caller The ASCII string of the calling function (ie "free",
 *               "realloc") for error printing.
 * 
 * @sideeffects Changes the permissions on the segment being freed to get the
 * 				info pointer.
 * @returns The heap entry info pointer corresponding to this memory segment.
 */
static struct heap_entry *nj_free_init(void *start, void *callstack, 
		char *caller)
{
	struct heap_entry *inf;
	TR;

	/* Simple check to hint us to the user trying to pass an address not
	 * returned by malloc */
	if((u_long)start & ~(NJ_PAGE_MASK))
	{
		__nj_eprintf("NJAMD/%s: Address passed (0x%lx) is not page-aligned\n", 
				caller, (u_long)start, __nj_align);
		raise(SIGSEGV);
		return NULL;
	}

	TR;

	/* Since this is strict underflow, we must unprotect the first page to
	 * access the segment length. */
	if(mprotect(start, NJ_PAGE_SIZE, PROT_READ) == -1)
	{   
		TR;                                                   
		__nj_perror(__FUNCTION__"/mprotect");
		raise(SIGSEGV);
		return NULL;
	}

	if(__nj_known_chk_free == ERROR_CHK_FREE)
	{
		TR;
		/* If the start pointer is NULL, this HAS to be a double free, cause
		 * the user couldn't have set this without a segfault */
		if(*(struct heap_entry **)start == 0)
		{
			__nj_eprintf("NJAMD/%s: Double %s of address 0x%lx\n", caller, caller, (u_long)start);
			raise(SIGSEGV);
			return NULL;

		}
		TR;
	}

	/* we know inf is valid, because it was protected */
	inf = *(struct heap_entry **)start;


	/* If there was room for the fencepost, check it's value */
	if(*(u_int *)(start+NJ_ALIGN_TO(inf->ulen, NJ_INT_ALIGNMENT)+NJ_PAGE_SIZE) != NJ_FENCEPOST)
	{
		__nj_eprintf("NJAMD/%s: heap corruption. Try using the overflow "
						"option to pinpoint source of error\n", caller);
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));
		__nj_dump_entry(inf, TRACE_DEPTH);
		raise(SIGSEGV);
		return NULL;
	}

	if(inf->u_addr.start != start)
	{
		__nj_eprintf("NJAMD/%s: Inconsistancy. Address passed (0x%lx) "
				"is not directly from malloc (or some pretty fierce heap "
			"corruption took place\n", caller, start);
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));
		__nj_dump_entry(inf, TRACE_DEPTH);
		raise(SIGSEGV);
		return NULL;
	}

	if(!NJ_ISSET(__nj_flags, NJ_NO_FREE_INFO))
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));

	TR;
	return inf;
}


/** 
 * Performs the second half of free/realloc. Remaps memory pointed to
 * by start as specified by the NJAMD_CHK_FREE setting.
 *
 * @param start The actual start of the mmaped region
 * @param inf The heap info pointer for this region
 * @sideeffects Stats are updated. Start becommes remmaped or assigned to the
 * 				free list.
 */
static void nj_free_fini(void *start, struct heap_entry *inf)
{
	/* start is acutally on the protected page. */
	int len = ROUND_TO_BLOCK(inf->ulen);
	int chk_free = __nj_known_chk_free;

	/* Check if allocated before we knew the protection scheme. So use the 
	 * preinit value (SEGV_CHK_FREE). */
	if((__nj_slab0 <= (u_long)start && (u_long)start < __nj_known_slab0))
		chk_free = PREINIT_CHK_FREE;

	/* Report the old memory as freed */
	__nj_update_stats(-inf->ulen, -ROUND_TO_BLOCK(inf->ulen), (chk_free == NO_CHK_FREE) ? -ROUND_TO_BLOCK(inf->ulen) : 0);

	/* If we don't keep freed info, return it to the heap-stack */
	if(NJ_ISSET(__nj_flags, NJ_NO_FREE_INFO))
		__nj_heap_stack_push(inf);
			
	switch(chk_free)
	{
		case ERROR_CHK_FREE:
			/* If error checking, map to the zero filled file */
			if(mmap(start, NJ_PAGE_SIZE, __nj_prot, MAP_FIXED | MAP_SHARED,
						__nj_zfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			if((len != NJ_PAGE_SIZE) && mmap(start+NJ_PAGE_SIZE, len-NJ_PAGE_SIZE, __nj_prot, 
						MAP_FIXED | MAP_PRIVATE | MAP_ANON, __nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		/* Unreliable for strict underflow */
		case SEGV_CHK_FREE:
			if(mmap(start, len, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		case NO_CHK_FREE:
			/* the start is protected. start+NJ_PAGE_SIZE is writeable where
			 * we can store the list */
			if(mprotect(start, NJ_PAGE_SIZE, PROT_NONE) == -1)
			{   
				TR;                                                   
				__nj_perror(__FUNCTION__"/mprotect");
				raise(SIGSEGV);
				return;
			}
			__nj_free_list_push(start, start+NJ_PAGE_SIZE, len);

#if 0	
			if(munmap(start, len) == -1)
			{
				__nj_perror(__FUNCTION__"/munmap");
			}
#endif
			break;

		case NO_FREE:
			/* Do nothing */
			break;

		default:
			__nj_eprintf("Unknown free checking option %d\n", chk_free);

			if(mmap(start, len, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* Will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;
	}

	
	TR;
}


/**
 * Provides allocated segments bounded by a protected page immediately before
 * the user's segment. This function handles malloc, calloc, and realloc. At
 * least two pages are allocated by this function, but unlike the standard
 * overflow and underflow, both pages must be faulted in order for us to be
 * able to save and recover the info pointer.
 *
 * @param ptr A previously allocated pointer for realloc
 * @param len The length to allocate
 * @param zero_fill If true, calloc
 * @param align Alignment. Not used since we are always aligned to a page. 
 * @param callstack The return addresses of the malloc call
 * @param caller The caller name, for error reporting.
 *
 * @sideeffects Free list is modified if NO_CHK_FREE. The heap table is modified 
 * 				(new info ptr). Updates NJAMD statistics.
 * @returns A mmaped, page-aligned, and protected segment.
 */
void *__nj_sunderflow_alloc(void *ptr, size_t len, int zero_fill, size_t align, 
		u_long *callstack, char *caller)
{
	struct heap_entry *inf;
	size_t fetch;
	void *buf = NULL;

	TR;
	if(!len && NJ_ISSET(__nj_flags, NJ_ALLOW_MALLOC0))
	{
		return NULL;
	}

	/* Ask a stupid question, get a stupid answer */
	if(len <= 0)
	{
		__nj_eprintf("NJAMD/%s: Request must be > 0 (%d)\n", caller, len);
		__nj_print_stack_trace(callstack, TRACE_DEPTH);
		errno = EINVAL;
		return NULL;
	}

	/* We need enough memory to fit an aligned length in, as well as the 
	 * fencepost that goes after it */
	fetch = ROUND_TO_BLOCK(NJ_ALIGN_TO(len, NJ_INT_ALIGNMENT) + sizeof(u_int));

	TR;
	/* Realloc? */
	if(ptr != NULL)
	{
		void *start = ptr - NJ_PAGE_SIZE;
		TR;	
		if(!(inf = nj_free_init(start, callstack, caller)))
		{
			/* This should not happen */
			return NULL;
		}
		
		TR;
		/* Get a chunk from the slab allocator */
		if(__nj_known_chk_free != NO_CHK_FREE ||
				(buf = __nj_free_list_pop(fetch)) == NULL)
			buf = __nj_slab_fetch(fetch);
		
		/* Copy the memory and info to the proper places */	
		memcpy(buf+NJ_PAGE_SIZE, ptr, NJ_MIN(inf->ulen, len));

		/* Remap the old memory to bupkes */
		nj_free_fini(start, inf);

		*(struct heap_entry **)buf = inf = __nj_new_heap_entry(buf, len, callstack);

		start = buf;
		TR;

	}
	else
	{
		TR;
		/* Get a chunk from the slab allocator */
		if(__nj_known_chk_free != NO_CHK_FREE ||
				(buf = __nj_free_list_pop(fetch)) == NULL)
			buf = __nj_slab_fetch(fetch);


		/* calloc */
		if(zero_fill)
		{
			int i;
			/* Page copies may me optimized by memcpy */
			for(i=1; i < (fetch >> NJ_PAGE_SHIFT); i++)
				memcpy(buf + (i<<NJ_PAGE_SHIFT), __nj_zero_page, NJ_PAGE_SIZE); 
		}

		TR;

		*(struct heap_entry **)buf = inf = __nj_new_heap_entry(buf, len, callstack);
	}

	/* Report the new memory */
	__nj_update_stats(len, fetch, fetch);

   /* Note that in the strict case, always protect the memory, because it
	* entered the free list unprotected.. */
	if(mprotect(buf, NJ_PAGE_SIZE, __nj_prot) == -1)
	{
		__nj_perror(__FUNCTION__"/mprotect");
		exit(1);
	}

	TR;

	/* To check for overflows */
	/* the fencepost is an int and must be aligned */
	*(u_int *)(buf+NJ_PAGE_SIZE+NJ_ALIGN_TO(len, NJ_INT_ALIGNMENT)) = NJ_FENCEPOST;

	return buf+NJ_PAGE_SIZE;
}

/** 
 * Main entry point for freeing memory. Basically just calls __nj_free_init
 * and __nj_free_fini. 
 *
 * @param buf Allocated buffer to free
 * @param callstack Return address call stack
 */ 
void __nj_sunderflow_free(void *buf, u_long *callstack)
{
	void *start = buf - NJ_PAGE_SIZE;
	struct heap_entry *inf;

	if(buf == NULL)
	{
		return;
	}

	TR;	
	if(!(inf = nj_free_init(start, callstack, "free")))
	{
		/* Shouldn't happen */
		TR;
		return;
	}
	
	nj_free_fini(start, inf);
}

// vim:ts=4
