/** 
 * @file mem_under.c 
 *
 * Provides the standard underflow capabilities of NJAMD
 *
 * These functions handle the case where a protected page of memory is placed
 * before info pointer that lies before the users segment. Thus there is no
 * need to fault the protected page, and this saves 4K per alloc. The info
 * pointer is checked for consistancy and validity during free.
 *
 * @FIXME: Consider consolidating this with mem_sunder.. or at least making it
 *         the mirror image.
 *
 * @NOTES: The layout of the memory is like this: protected page, inf pointer
 * aligned to a requested alignment, user buffer. inf->ulen is not aligned in
 * any way. It is the user's requested length.
 * 
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 */
#include <lib/njamd.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <errno.h>



/**
 * Performs the first half of free/realloc. This includes retrieving an info
 * pointer from the segment, as well as checking fencepost values.
 *
 * @param start The first pagesize boundary below userbuf
 * @param buf The user's buffer (start + aligned NJ_ACCOUNTING_OVERHEAD)
 * @param callstack The return addresses callstack
 * @param caller The ASCII string of the calling function (ie "free",
 *               "realloc") for error printing.
 * 
 * @sideeffects In some cases the permissions on start change to readonly.
 * @returns The heap entry info pointer corresponding to this memory segment.
 */
static struct heap_entry *nj_free_init(void *start, void *buf, void *callstack,
		char *caller)
{
	struct heap_entry *inf;
	size_t align_offset = buf - start;
	TR;

	if(__nj_known_chk_free == ERROR_CHK_FREE)
	{
		/* If this is error check free, we want to try to determine if this
		 * memory was freed already */
		if(mprotect(start, NJ_PAGE_SIZE, PROT_READ) == -1)
		{
			TR;
			__nj_perror(__FUNCTION__"/mprotect");
			raise(SIGSEGV);
			return NULL;
		}

		TR;
		if(*(struct heap_entry **)start == 0)
		{
			/* if the size was zero, but the whole page isn't, we have an
 			 * underflow of 0.*/
			if(memcmp(start, __nj_zero_page, NJ_PAGE_SIZE) != 0)
			{
				TR;
				__nj_eprintf("NJAMD: heap corruption. Inconsistant allocation "
						"sizes. Try using the strict underflow option to " 
						" pinpoint source of error\n");	
				raise(SIGSEGV);
				return NULL;
			}
			__nj_eprintf("NJAMD/%s: Double %s of address 0x%lx\n"
					"0x%lx\n", caller, caller, (u_long)buf);
			raise(SIGSEGV);
			return NULL;

		}
		TR;
	}

	inf = *(struct heap_entry **)start;

	if(!__nj_is_valid_heap_ptr(inf))
	{
		__nj_eprintf("NJAMD/%s: Heap corruption detected on %s of 0x%lx.\n\tTry using strict underflow option\n",
				caller, caller, (u_long)buf);
		raise(SIGSEGV);
		exit(1);
	}

	/* NJ_FENCEPOST is an int, and must lie at an aligned boundary independant
	 * of what the user wants. Buf can be aligned to anything, so we have to
	 * go from start */
	if(*(u_int *)(start + NJ_ALIGN_TO(inf->ulen + align_offset, NJ_INT_ALIGNMENT)) != NJ_FENCEPOST)
	{
		__nj_eprintf("NJAMD/%s: heap corruption. Try using the overflow "
						"option to pinpoint source of error\n", caller);
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));
		__nj_dump_entry(inf, TRACE_DEPTH);
		raise(SIGSEGV);
		return NULL;
	}

	if(inf->u_addr.start != start - NJ_PAGE_SIZE)
	{
		__nj_eprintf("NJAMD/%s: Inconsistancy. Either address passed (0x%lx) "
				"is not directly from malloc, or small underflow. Try using "
				"strict underflow option. \n", caller, buf);
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));
		__nj_dump_entry(inf, TRACE_DEPTH);
		raise(SIGSEGV);
		return NULL;
	}

	if(!NJ_ISSET(__nj_flags, NJ_NO_FREE_INFO))
		memcpy(inf->loc.freed, callstack, TRACE_DEPTH*sizeof(u_long));

	TR;
	return inf;
}

/** 
 * Performs the second half of free/realloc. Remaps memory pointed to
 * by start as specified by the NJAMD_CHK_FREE setting.
 *
 * @param start The start of unprotected memory (ie 1 page into the mapped
 * 				segment)
 * @param inf The heap info pointer for this region
 * @param align_offset the offset between the start and the user's buf
 * @sideeffects Stats are updated. Start becommes remmaped or assigned to the
 * 				free list.
 */
static void nj_free_fini(void *start, size_t align_offset, 
		struct heap_entry *inf) 
{
	/* start is acutally AFTER the protected page. */
	int len = ROUND_TO_BLOCK(inf->ulen + align_offset) - NJ_PAGE_SIZE;
	int chk_free = __nj_known_chk_free;
	

	/* Check if allocated before we knew the protection scheme. So use the 
	 * preinit value (SEGV_CHK_FREE). */
	if((__nj_slab0 <= (u_long)start && (u_long)start < __nj_known_slab0))
		chk_free = PREINIT_CHK_FREE;

	/* We pass in negative arguments to subtract that lengh. But we only
	 * subtract the address space if NO_CHK_FREE is set. */
	__nj_update_stats(-inf->ulen, 
			-(ROUND_TO_BLOCK(inf->ulen + align_offset) - NJ_PAGE_SIZE), 
			(chk_free == NO_CHK_FREE) ? -ROUND_TO_BLOCK(inf->ulen + align_offset) : 0); 

	/* If we don't keep freed info, return it to the heap-stack */
	if(NJ_ISSET(__nj_flags, NJ_NO_FREE_INFO))
		__nj_heap_stack_push(inf);
	
	switch(chk_free)
	{
		case ERROR_CHK_FREE:
			/* start-PAGE_SIZE is already protected. */
			if(mmap(start, NJ_PAGE_SIZE, __nj_prot, MAP_FIXED | MAP_SHARED,
						__nj_zfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			if((len != NJ_PAGE_SIZE) && mmap(start+NJ_PAGE_SIZE, len-NJ_PAGE_SIZE, __nj_prot, 
						MAP_FIXED | MAP_PRIVATE | MAP_ANON, __nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		case SEGV_CHK_FREE:
			if(mmap(start, len, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;

		case NO_CHK_FREE:
			TR;
			/* start-PAGE_SIZE is protected already, so it's not even passed
			 * to this fcn. start is writeable where we can store the list. 
			 * len, if you'll notice, has a page_size subtracted from it when
			 * defined.*/
			__nj_free_list_push(start-NJ_PAGE_SIZE, start, len+NJ_PAGE_SIZE);
#if 0	
			if(munmap(start-NJ_PAGE_SIZE, len+NJ_PAGE_SIZE) == -1)
			{
				__nj_perror(__FUNCTION__"/munmap");
			}
#endif
			break;

		case NO_FREE:
			/* Do nothing */
			break;

		default:
			__nj_eprintf("Unknown free checking option %d\n", chk_free);

			if(mmap(start, len, __nj_prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON,
						__nj_anonfd, 0) == (void *)-1)
			{
				if(errno == ENOMEM)
					__nj_out_of_memory(); /* will exit */
				else
					__nj_perror(__FUNCTION__"/mremap");
			}
			break;
	}

	TR;
}

/**
 * Provides allocated segments bounded by a protected page before
 * the user's segment. This function handles malloc, calloc, and realloc. At
 * least two pages are allocated by this function, but the info pointer is
 * stored AFTER the protected page, yet before the user segment, thus there is
 * no need to fault the protected page, and this saves 4K per alloc over the
 * strict underflow method.
 *
 * @param ptr A previously allocated pointer for realloc
 * @param len The length to allocate
 * @param zero_fill If true, calloc
 * @param align The alignment to use on this buffer
 * @param callstack The return addresses of the malloc call, used for accounting
 * 				purposes.
 * @param caller The caller name, for error reporting.
 *
 * @sideeffects Free list is modified if NO_CHK_FREE. The heap table is modified 
 * 				(new info ptr). Updates NJAMD statistics.
 * @returns A mmaped, aligned, and protected segment.
 */
void *__nj_underflow_alloc(void *ptr, size_t len, int zero_fill, size_t align,
		u_long *callstack, char *caller)
{
	struct heap_entry *inf;
	size_t fetch;
	int protect_mem = FALSE;
	void *buf = NULL;

	TR;
	if(!len && NJ_ISSET(__nj_flags, NJ_ALLOW_MALLOC0))
	{
		return NULL;
	}

	/* Ask a stupid question, get a stupid answer */
	if(len <= 0)
	{
		__nj_eprintf("NJAMD/%s: Request must be > 0 (%d)\n", caller, len);
		__nj_print_stack_trace(callstack, TRACE_DEPTH);
		errno = EINVAL;
		return NULL;
	}

	/* We need enough memory to fit len, and the inf pointer, aligned to align,
	 * as well as the fencepost after that whole mess is aligned to
	 * NJ_INT_ALIGNMENT */
	fetch = ROUND_TO_BLOCK(NJ_ALIGN_TO(len + NJ_ALIGN_TO(NJ_ACCOUNTING_OVERHEAD, align), NJ_INT_ALIGNMENT) + sizeof(u_int));

	TR;
	/* Realloc? */
	if(ptr != NULL)
	{
		void *start = (void *)((u_long)ptr & NJ_PAGE_MASK);
		TR;	
		if(!(inf = nj_free_init(start, ptr, callstack, caller)))
		{
			/* This should not happen */
			return NULL;
		}

		TR;
	  	
		if(__nj_known_chk_free != NO_CHK_FREE ||
				(buf = __nj_free_list_pop(fetch)) == NULL)
		{
			buf = __nj_slab_fetch(fetch);
			/* protect mem that comes from the slab */
			protect_mem = TRUE;
		}
		
		/* Copy the memory and info to the proper places */	
		memcpy(buf+NJ_ALIGN_TO(NJ_ACCOUNTING_OVERHEAD, align)+NJ_PAGE_SIZE, ptr, NJ_MIN(inf->ulen, len));

		/* Remap the old memory to bupkes */
		nj_free_fini(start, (u_long)ptr-(u_long)start, inf);

		*(struct heap_entry **)(buf + NJ_PAGE_SIZE) = inf = __nj_new_heap_entry(buf, len, callstack);

		start = buf;
		TR;
	}
	else
	{
		TR;
		if(__nj_known_chk_free != NO_CHK_FREE ||
				(buf = __nj_free_list_pop(fetch)) == NULL)
		{
			buf = __nj_slab_fetch(fetch);
			/* protect mem that comes from the slab */
			protect_mem = TRUE;
		}

		/* calloc */
		if(zero_fill)
		{
			int i;
			/* Dont fault that first page */
			for(i=1; i < (fetch >> NJ_PAGE_SHIFT); i++)
				memcpy(buf + (i<<NJ_PAGE_SHIFT), __nj_zero_page, NJ_PAGE_SIZE); 
		}
		TR;

		*(struct heap_entry **)(buf+NJ_PAGE_SIZE) = inf = __nj_new_heap_entry(buf, len, callstack);
	}

	__nj_update_stats(len, fetch-NJ_PAGE_SIZE, fetch);
		 
	if(protect_mem)
		if(mprotect(buf, NJ_PAGE_SIZE, __nj_prot) == -1)
		{
			__nj_perror(__FUNCTION__"/mprotect");
			exit(1);
		}

	TR;

	/* To check for overflows */	
	/* NJ_FENCEPOST is an int, and must lie at a boundary specified by
	 * this processor, not the user specified alignment */
	*(u_int *)(buf+NJ_PAGE_SIZE+NJ_ALIGN_TO(len+NJ_ALIGN_TO(NJ_ACCOUNTING_OVERHEAD, align), NJ_INT_ALIGNMENT)) = NJ_FENCEPOST;

	return buf + NJ_PAGE_SIZE + NJ_ALIGN_TO(NJ_ACCOUNTING_OVERHEAD, align);
}

/** 
 * Main entry point for freeing memory. Basically just calls __nj_free_init
 * and __nj_free_fini. 
 *
 * @param buf Allocated buffer to free
 * @param ret Return address call stack
 */ 
void __nj_underflow_free(void *buf, u_long *callstack)
{
	/* The buffer starts an arbitarty alignment before buf on a page boundry.
	 * See about 15 lines above. */
	void *start = (void *)((u_long)buf & NJ_PAGE_MASK);
	struct heap_entry *inf;

	if(buf == NULL)
	{
		return;
	}

	TR;	
	if(!(inf = nj_free_init(start, buf, callstack, "free")))
	{
		/* Shouldn't happen */
		TR;
		return;
	}

	nj_free_fini(start, (u_long)buf - (u_long)start, inf);
}

// vim:ts=4
