/** 
 * @file public.c
 * 
 * Provides the libc interface into NJAMD.
 *
 * This is how we override malloc, realloc and calloc. The functions in NJAMD
 * take precedence over libc when preloaded.
 *
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 * 
 */

/* Avoid conflicting types */
#define pthread_create __nj_dummy1 
#define pthread_exit __nj_dummy2 
#define signal __nj_dummy4
#define exit __nj_dummy5

#include <lib/njamd.h>
#include <lib/libc.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>

#undef exit
#undef signal
#undef pthread_exit
#undef pthread_create

#ifdef HAVE_SANE_DLOPEN
/**
 * Wraps libc signal, makes sure the user doesn't overwrite singals NJAMD
 * needs to handle. 
 *
 * NJAMD needs to catch segv, bus error and other deadly signals to clean up
 * properly and display some info to the user, as well as USR1 to dump leaks. 
 * So we wrap signal to save the handler the user wants to execute, and then
 * when the signal actually occurs we just call their functions.
 *
 * @param num signal number
 * @param newhandler New signal handler function
 * @returns The previous signal handler the user may have registered.
 * @sideeffects Replaces the signal table entry at sig with the newhandler.
 *              More importantly, the #__nj_sig_actn action table entry is set
 *              to NULL for this signal.  Also runs the init functions if they
 *              haven't already.
 * @see __nj_sig_dispatch()
 */
nj_sighandler_t signal(int num, nj_sighandler_t newhandler)
{
	nj_sighandler_t ret;
	TR;

	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}
	TR;

# ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
# endif

	if(!NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY) 
			&& NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		__nj_secondary_init();
	}
	TR;

	/* If we dump hardcore, we don't handle segv */
	if(NJ_ISSET(__nj_flags, NJ_DUMP_HARDCORE))
	{
		switch(num)
		{
			case SIGQUIT:
			case SIGABRT:
			case SIGUSR1:
			case SIGINT:
			case SIGTERM:
				/* Look up ret in the handler array, save the new one, and
				 * declaire no sigaction for it. return ret */
				ret = __nj_sig_hdlr[num];
				__nj_sig_hdlr[num] = newhandler;
				__nj_sig_actn[num] = NULL;
				return ret;
		}

	}
	else
	{
		switch(num)
		{
			case SIGQUIT:
			case SIGILL:
			case SIGABRT:
			case SIGFPE:
			case SIGUSR1:
			case SIGSEGV:
			case SIGBUS:
			case SIGINT:
			case SIGTERM:
				/* Look up ret in the handler array, save the new one, and
				 * declaire no sigaction for it. return ret */
				ret = __nj_sig_hdlr[num];
				__nj_sig_hdlr[num] = newhandler;
				__nj_sig_actn[num] = NULL;
				return ret;
		}
	}
	TR;
	/* If we are still here, we don't care about the signal. Just let the 
	 * kernel handle it 
	 */
	if(__njLibcSignal)
		return __njLibcSignal(num, newhandler);
	else
	{
		__nj_eprintf("Error!!! Undefined symbol signal\n");
		errno = ENOSYS;
		return NULL;
	}
}

# ifdef HAVE_POSIX_SIGACTION
/**
 * Wraps libc sigaction, makes sure the user doesn't overwrite singals NJAMD
 * needs to handle. 
 *
 * NJAMD needs to catch segv, bus error and other deadly signals to clean up
 * properly and display some info to the user, as well as USR1 to dump leaks. 
 * So we wrap signal to save the handler the user wants to execute, and then
 * when the signal actually occurs we just call their functions.
 *
 * @param num signal number
 * @param new New sigaction descriptor.
 * @param old Old sigaction descriptor is stored here if non-null
 * @returns -1 On error.
 * @sideeffects Replaces the signal table entry at sig with the new action.
 *              More importantly, the __nj_sig_hdlr table entry is set
 *              to NULL for this signal if we are using SA_SIGINFO.
 *              Also runs the init functions if they haven't already.
 * @see __nj_sig_dispatch()
 */
int sigaction(int num, const struct sigaction *new, struct sigaction *old)
{
	
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}
	TR;

#  ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#  endif

	if(!NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY) 
			&& NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		__nj_secondary_init();
	}
	TR;
	
	/* Handle deadly signals internally, so we always exit cleanly and with
	 * status messages, if the user wishes */
	if(NJ_ISSET(__nj_flags, NJ_DUMP_HARDCORE))
	{
		switch(num)
		{
			case SIGUSR1:
			case SIGQUIT:
			case SIGABRT:
			case SIGINT:
			case SIGTERM:
				TR;
				if(old)
				{
					old->sa_handler = __nj_sig_hdlr[num];
					old->sa_sigaction = __nj_sig_actn[num];
				}
				if(new)
				{
					if(new->sa_flags & SA_SIGINFO)
					{
						__nj_sig_actn[num] = new->sa_sigaction;
						__nj_sig_hdlr[num] = SIG_DFL;
					}
					else
					{
						__nj_sig_hdlr[num] = new->sa_handler;
						__nj_sig_actn[num] = NULL;
					}
				}
				return 0;
		}
	}
	else
	{
		switch(num)
		{
			case SIGUSR1:
			case SIGILL:
			case SIGFPE:
			case SIGQUIT:
			case SIGABRT:
			case SIGSEGV:
			case SIGBUS:
			case SIGINT:
			case SIGTERM:
				TR;
				if(old)
				{
					old->sa_handler = __nj_sig_hdlr[num];
					old->sa_sigaction = __nj_sig_actn[num];
				}
				if(new)
				{
					if(new->sa_flags & SA_SIGINFO)
					{
						__nj_sig_actn[num] = new->sa_sigaction;
						__nj_sig_hdlr[num] = SIG_DFL;
					}
					else
					{
						__nj_sig_hdlr[num] = new->sa_handler;
						__nj_sig_actn[num] = NULL;
					}
				}
				return 0;
		}
	}
	TR;
	if(__njLibcSigaction)
		return __njLibcSigaction(num, new, old);
	else
	{
		__nj_eprintf("Error!!! Undefined symbol sigaction\n");
		errno = ENOSYS;
		return -1;
	}
}
# endif /* SIGACTION */

# ifdef _THREAD_SAFE
/**   
 * We wrap pthread_create so we KNOW that __nj_primary_init() will run
 * before any threads exist. Also, in some threads implementations, the bottom
 * of the stack isn't well defined (ie, it's just some garbage pointer). So we
 * need to start all threads from a known function, and then stop backtrace at
 * that function (__nj_pthread_launch())
 * 
 * @par Paramaters: Those of pthread_create(3)
 * @sideeffects Calls __nj_primary_init() if not called already. This way we know
 * 				that all init functions run before we need mutexes to protect
 * 				against race conditions.
 */
int  pthread_create(pthread_t  *  thread, pthread_attr_t *attr,
		void * (*start_routine)(void *), void * arg)
{
	static struct pthread_arg_wrapper warg;
	int ret;

	TR;
	NJ_SET(__nj_flags, NJ_ACTIVE_PTHREADS);

	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}
	TR;

#  ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#  endif

	if(!NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY) && NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		__nj_secondary_init();
	}
	TR;
	
	/* we need this lock to protect the static variable, which must be static
	 * because this function most likely will return before pthread_launch is
	 * called. The lock is released by __nj_pthread_launch */
	pthread_mutex_lock(&__nj_pthread_lock);

	warg.start = start_routine;
	warg.real_arg = arg;
	
	if(__nj_pthread_create)
		ret = __nj_pthread_create(thread, attr, __nj_pthread_launch, &warg);
	else
	{
		__nj_eprintf("Error!!! Undefined symbol pthread_create\n");
		ret = -1;
	}

   /* This unlock moved to __nj_pthread_launch to fix a race condition spotted
    * by Aleks Totic
	pthread_mutex_unlock(&__nj_pthread_lock); */

	return ret;
}

/** 
 * Food for the dynamic linker. Do NOT move. 
 * Used to determine the start of pthread_exit so that we can tell if a
 * descructor was called during exit of a thread. 
 */
void __nj_pthread_exit_start()
{
	return;
}

/** 
 * Simply wraps pthread_exit. We need to be able to find the end of the stack
 * if a destructor was called.
 */
void pthread_exit(void *retval)
{

	/* Pthread exit damn well better not be called before create */
	if(__nj_pthread_exit)
	{
		__nj_pthread_exit(retval);
	}
	else
	{
		__nj_eprintf("Error!!! Undefined symbol pthread_exit\n");
	}

}
# endif /* THREADSAFE */
#endif /* DLOPEN */

/** Do NOT move this from pthread_exit. Used to determine it's end value */
void __nj_pthread_exit_end(void)
{
	return;
}

#ifdef HAVE_SANE_DLOPEN

/** Food for the dynamic linker. Do NOT move. Same thing. Used to determine
 * the end of the stack for destructors */
void __nj_exit_start(void)
{
	return;
}

/**
 * Wrapped exit, used just to determine the end of stack when free is called
 * from a destructor. 
 */
void exit(int ret)
{
	if(!NJ_ISSET(__nj_flags, NJ_RAN_LIBC))
	{
		__njLibcFuncInit();
	}
	
	if(__nj_exit)
	{
		__nj_exit(ret);
	}

	__nj_eprintf("Error!!! Undefined symbol exit\n");
	nj_exit(ret);
	
}
#endif

/** Do NOT move this from exit() */
void __nj_exit_end(void)
{
	return;
}

/**
 * The main library entry point. It calls the
 * appropriate alloc via the #__nj_known_alloc pointer, if we have run
 * __nj_secondary_init(). If secondary init has NOT run, use the default preinit
 * allocator.
 *
 * @param size The size to allocate
 * @returns The allocated segment
 * @sideeffects Calls the init functions if they haven't been run, and it's
 * 				time (ie __nj_secondary_init()) is only called if main has
 * 				been entered).
 * @see __nj_none_alloc() __nj_overflow_alloc() __nj_sunderflow_alloc() __nj_underflow_alloc()
 */ 
void *malloc(size_t size)
{
	u_long ret_addrs[TRACE_DEPTH];

	TR;
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}

	TR;

#ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#endif

	__nj_userspace_ret(ret_addrs, TRACE_DEPTH);

	/* If the allocator type is unknown (ie the seconday 
	   init hasn't been run yet), use the default one */
	if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY))
	{
		TR;
		return __nj_known_alloc(NULL, size, 0, __nj_align, ret_addrs, __FUNCTION__);
	}
	else if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		TR;
		return PREINIT_ALLOCATOR(NULL, size, 0, __nj_align, ret_addrs, __FUNCTION__);
	}
	else /* we are in main, but didn't run init yet */
	{
		TR;
		__nj_secondary_init();
		return __nj_known_alloc(NULL, size, 0, __nj_align, ret_addrs, __FUNCTION__);
	}
}

/**
 * Implements valloc, which simply calls malloc.
 *
 * @param size The size to allocate
 * @returns The allocated segment
 * @sideeffects Calls the init functions if they haven't been run, and it's
 * 				time (ie __nj_secondary_init() is only called if main has been entered).
 * @see __nj_none_alloc() __nj_overflow_alloc() __nj_sunderflow_alloc() __nj_underflow_alloc()
 */ 
void *valloc(size_t size)
{
	return malloc(size);
}

/**
 * Implements memalign, to returned aligned memory. It calls the
 * appropriate alloc via the __nj_known_alloc pointer, if we have run
 * __nj_secondary_init(). If secondary init has NOT run, use the default preinit
 * allocator. The only difference between this and malloc is that the
 * alignment is passed to the allocation function
 * 
 * @param alignment The requested alignment
 * @param size The size to allocate
 * @returns The allocated segment
 * @sideeffects Calls the init functions if they haven't been run, and it's
 * 				time (ie secondary_init is only called if main has been
 * 				entered).
 * @see __nj_none_alloc() __nj_overflow_alloc() __nj_sunderflow_alloc() __nj_underflow_alloc()
 */
void *memalign(size_t alignment, size_t size)
{
	u_long ret_addrs[TRACE_DEPTH];

	TR;
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}

	TR;

#ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#endif

	__nj_userspace_ret(ret_addrs, TRACE_DEPTH);

	/* If the allocator type is unknown (ie the seconday 
	   init hasn't been run yet), use the default one */
	if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY))
	{
		TR;
		return __nj_known_alloc(NULL, size, 0, alignment, ret_addrs, __FUNCTION__);
	}
	else if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		TR;
		return PREINIT_ALLOCATOR(NULL, size, 0, alignment, ret_addrs, __FUNCTION__);
	}
	else /* we are in main, but didn't run init yet */
	{
		TR;
		__nj_secondary_init();
		return __nj_known_alloc(NULL, size, 0, alignment, ret_addrs, __FUNCTION__);
	}
	
}

/**
 * Implements posix_memalign, to returned aligned memory. It calls the
 * appropriate alloc via the __nj_known_alloc pointer, if we have run
 * __nj_secondary_init(). If secondary init has NOT run, use the default preinit
 * allocator. The only difference between this and realloc is that the
 * alignment is passed to the allocation function.
 * 
 * @param alignment The requested alignment 
 * @param size The size to allocate
 * @returns I have no idea. Does anyone have a POSIX spec handy?
 * @sideeffects Calls the init functions if they haven't been run, and it's
 * 				time (ie secondary_init is only called if main has been
 * 				entered).
 * @see __nj_none_alloc() __nj_overflow_alloc() __nj_sunderflow_alloc() __nj_underflow_alloc()
 */
int posix_memalign(void **buf, size_t alignment, size_t size)
{
	u_long ret_addrs[TRACE_DEPTH];
	
	TR;	
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}

	TR;

#ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#endif

	if(!buf)
	{
		__nj_eprintf("NJAMD/posix_memalign: NULL pointer passed\n");
		return -1;
	}
	
	__nj_userspace_ret(ret_addrs, TRACE_DEPTH);
	
	if(*buf && !__nj_address_from_malloc(*buf))
	{
		__nj_eprintf("NJAMD/free: Address %p not from malloc\n", buf);
		__nj_print_stack_trace(ret_addrs, TRACE_DEPTH);
		return -1;
	}


	if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY))
	{
known:
		if(size != 0)
		{
			*buf = __nj_known_alloc(*buf, size, 0, alignment, ret_addrs, __FUNCTION__);
			return 0;
		}
		else
		{
			/* if ptr is from a section of memory allocated before we knew which
			 * allocator to use, free it with default */
			if(!((u_long)*buf >= __nj_slab0 && (u_long)*buf < __nj_known_slab0))
				__nj_known_free(*buf, ret_addrs);
			else
				PREINIT_DEALLOCATOR(*buf, ret_addrs);
			return 0;
		}
	}
	else if(NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		__nj_secondary_init();
		goto known;
	}

	if(size != 0)
	{
		/* must write realloc pair using mremap */
		*buf = PREINIT_ALLOCATOR(*buf, size, 0, alignment, ret_addrs, __FUNCTION__);
		return 0;
	}
	else 
	{
		PREINIT_DEALLOCATOR(*buf, ret_addrs);
		return 0;
	}
	
}


/**
 * The main library entry point. It calls the
 * appropriate alloc via the #__nj_known_alloc pointer, if we have run
 * __nj_secondary_init(). If secondary init has NOT run, use the default preinit
 * allocator. The only difference between this and malloc is that zero_fill is
 * set to true.
 *
 * @param size The size to allocate
 * @returns The allocated segment
 * @sideeffects Calls the init functions if they haven't been run, and it's
 * 				time (ie secondary_init is only called if main has been entered).
 * @see __nj_none_alloc() __nj_overflow_alloc() __nj_sunderflow_alloc() __nj_underflow_alloc()
 */ 
void *calloc(size_t nmemb, size_t size)
{
	u_long ret_addrs[TRACE_DEPTH];

	TR;
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}

	TR;
#ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#endif
	
	__nj_userspace_ret(ret_addrs, TRACE_DEPTH);

	if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY))
	{
		return __nj_known_alloc(NULL, size*nmemb, 1, __nj_align, ret_addrs, __FUNCTION__);
	}
	else if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		return PREINIT_ALLOCATOR(NULL, size*nmemb, 1, __nj_align, ret_addrs, __FUNCTION__);
	}
	else /* we are in main, but didn't run init yet */
	{
		__nj_secondary_init();
		return __nj_known_alloc(NULL, size*nmemb, 1, __nj_align, ret_addrs, __FUNCTION__);
	}
	
}

/**
 * Implements realloc, calling free if size is 0. It calls the
 * appropriate alloc via the __nj_known_alloc pointer, if we have run
 * __nj_secondary_init(). If secondary init has NOT run, use the default preinit
 * allocator. The only difference between this and malloc is that the ptr is
 * passed to the allocation function
 * 
 * @param ptr the previously allocated function.
 * @param size The size to allocate
 * @returns The allocated segment
 * @sideeffects Calls the init functions if they haven't been run, and it's
 * 				time (ie secondary_init is only called if main has been
 * 				entered).
 * @see __nj_none_alloc() __nj_overflow_alloc() __nj_sunderflow_alloc() __nj_underflow_alloc()
 */
void *realloc(void *ptr, size_t size)
{
	u_long ret_addrs[TRACE_DEPTH];
	
	TR;	
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		__nj_primary_init();
	}

	TR;

#ifndef HAVE_ATTRIB_CONSTRUCT
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN) && __nj_entered_usercode())
	{
		NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	}
#endif

	__nj_userspace_ret(ret_addrs, TRACE_DEPTH);

	if(ptr && !__nj_address_from_malloc(ptr))
	{
		__nj_eprintf("NJAMD/free: Address %p not from malloc\n", ptr);
		__nj_print_stack_trace(ret_addrs, TRACE_DEPTH);
		return NULL;
	}

	if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY))
	{
known:
		if(size != 0)
		{
			return __nj_known_alloc(ptr, size, 0, __nj_align, ret_addrs, __FUNCTION__);
		}
		else
		{
			/* if ptr is from a section of memory allocated before we knew which
			 * allocator to use, free it with default */
			if(!((u_long)ptr >= __nj_slab0 && (u_long)ptr < __nj_known_slab0))
				__nj_known_free(ptr, ret_addrs);
			else
				PREINIT_DEALLOCATOR(ptr, ret_addrs);
			return NULL;
		}
	}
	else if(NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN))
	{
		__nj_secondary_init();
		goto known;
	}

	if(size != 0)
	{
		/* must write realloc pair using mremap */
		return PREINIT_ALLOCATOR(ptr, size, 0, __nj_align, ret_addrs, __FUNCTION__);
	}
	else 
	{
		PREINIT_DEALLOCATOR(ptr, ret_addrs);
		return NULL;
	}

}


/**
 * Implements libc free. We keep track of the region of memory allocated
 * with an unknown allocator by looking at the slab sections set up in
 * __nj_secondary_init().
 *
 * @param ptr The pointer to free.
 * @see __nj_none_free() __nj_overflow_free() __nj_sunderflow_free() __nj_underflow_free()
 */
void free(void *ptr)
{
	u_long ret_addrs[TRACE_DEPTH];

	if(!ptr && NJ_ISSET(__nj_flags, NJ_ALLOW_FREE0))
	{
		return;
	}
	
	__nj_userspace_ret(ret_addrs, TRACE_DEPTH);

	TR;
	if(ptr && __nj_address_from_malloc(ptr))
	{
		TR;
		/* if ptr is from a section of memory allocated before we knew which
		 * allocator to use, free it with default. */
		if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY) 
				&& !(__nj_slab0 <= (u_long)ptr && (u_long)ptr < __nj_known_slab0))
		{
			TR;
			__nj_known_free(ptr, ret_addrs);
		}
		else /* We don't run init from free. It's not nessicary */
			PREINIT_DEALLOCATOR(ptr, ret_addrs);
	}
	else if (ptr)
	{
		__nj_eprintf("NJAMD/free: Address %p not from malloc\n", ptr);
		__nj_print_stack_trace(ret_addrs, TRACE_DEPTH);
		return;
	}
}
// vim:ts=4
