/** @file setup.c initialization routines.
 *
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 * 
 */
#include <netdb.h> /* Mother f*#@ing piece of sh*@ AIX.. */
#include <lib/njamd.h>
#include <lib/libc.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>

/*@{ @name Internal debugging symbols*/
/** 
 *  Since we can't use gdb on PRELOADED libraries, we have to do some 
 *  creative bookkeaping. The TR define saves __FUNCTION__ and __LINE__
 *  to these variables, so our signal handler can spit them out
 */
#if defined(DEBUG) || defined(EDEBUG)
char					*file = "libc";
char 					*func = "_init";
int 					line;
#endif
/*@}*/

/**@{ @name global mm data structures */
struct slab_tbl 		__nj_slab_tbl;
struct free_list		*__nj_free_tbl[4];
struct heap_entry 		*__nj_heap_stack;
struct heap_tbl	 		__nj_heap_tbl;
struct heap_mid_dir		__nj_heap_mid_dir;
struct heap_top_dir		__nj_heap_top_dir;
#ifdef _THREAD_SAFE
pthread_mutex_t 		__nj_pthread_lock;
pthread_mutex_t 		__nj_heap_lock;
pthread_mutex_t 		__nj_slab_lock;
pthread_mutex_t 		__nj_init_lock;
pthread_mutex_t 		__nj_list_lock;
#endif
int 					__nj_zfd;
int 					__nj_efd = STDERR_FILENO;
#ifdef NJ_PAGE_SIZE
u_char 					__nj_zero_page[NJ_PAGE_SIZE];
#else
u_char					*__nj_zero_page;
#endif
int						__nj_anonfd = -1;
/*@}*/

/**@{ @name Allocator functions */
nj_allocator_t			__nj_known_alloc = DEFAULT_ALLOCATOR;
nj_deallocator_t		__nj_known_free = DEFAULT_DEALLOCATOR;
/*@}*/

/**@{ @name Prefrence variables */
int 					__nj_known_chk_free = DEFAULT_CHK_FREE;
u_long 					__nj_align = NJ_ALIGNMENT;
int						__nj_prot = PROT_NONE;
int						__nj_display_depth;
/*@}*/

/**@{ @name Status variables */
char			 		__nj_heap_file[80];
u_long					__nj_known_slab0;
u_long					__nj_sbrk0;
u_long					__nj_slab0;
struct mem_stats		__nj_mem_stats;
u_int 					__nj_leaked_user;
u_int 					__nj_leaked_faulted;
/*@}*/

/** This is where the boolean environment settings are kept */
u_int					__nj_flags;

/**@{ @name old signal handlers 
 * @see __nj_sig_dispatch() signal() sigaction() 
 */
nj_sighandler_t			__nj_sig_hdlr[sizeof(u_long)*8];
#ifdef HAVE_POSIX_SIGACTION
nj_sigaction_t			__nj_sig_actn[sizeof(u_long)*8];
#endif
/*@}*/

/* These variables are defined for portability, if the PAGE_SIZE and friends
 * aren't available as compile time constants
 */
#ifndef NJ_PAGE_SIZE
# if defined(DEBUG) || defined(EDBUG)
#  warning "**** Page Size undefined ****"
# endif
u_long 					NJ_PAGE_SIZE;
#endif

#ifndef NJ_PAGE_MASK
u_long 					NJ_PAGE_MASK;
#endif

#ifndef NJ_PAGE_SHIFT
u_long 					NJ_PAGE_SHIFT;
#endif


static void __nj_startup() __attribute__ ((constructor));
static void __nj_cleanup() __attribute__ ((destructor));


/**
 * Indirectly marks when main() starts.
 * 
 * This function guesses that by the time it was called, the program is just
 * about to enter main, since it is given a constructor attribute. It also
 * calls the primary NJAMD init function.
 */
static void __nj_startup()
{
	/* Some archs (*BSD) don't call malloc before entering main at all. So we
	 * want to run primary init now before any threads exist */
#ifdef HAVE_ATTRIB_CONSTRUCT
	TR;
	if(!NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
		__nj_primary_init();
	NJ_SET(__nj_flags, NJ_ENTERED_MAIN);
	TR;
#endif
}

/**
 * Handles proper shutdown of an NJAMD program.
 * 
 * All attempts are made to ensure this function is called no matter what on
 * exit. It handles dumping of leak info, displaying stats, and syncing and
 * truncating the heap file on disk.
 * 
 * @sideeffects Saves the state of the statistics to the heap file, and
 * destroys ALL pthread mutexes.
 */
static void __nj_cleanup()
{
	TR;
	/* This check for dlopen is done cause we can detect exit ourselves if we
	 * have dlopen to fetch the exit and pthread_exit symbols. If we don't
	 * have dlopen, we have to set a flag to guess when destructors start
	 * exiting */
#ifndef HAVE_SANE_DLOPEN
	NJ_SET(__nj_flags, NJ_ENTERED_DESTRUCT);
#endif

	if(__nj_display_depth) 
		__nj_dump_leaks(__nj_display_depth);

	if(NJ_ISSET(__nj_flags, NJ_DUMP_STATS))
		__nj_usage_stats();

	if(!NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY) 
			&& !NJ_ISSET(__nj_flags, NJ_PERSIST_HEAP))
		unlink(__nj_heap_file);

	/* We can't unmap this yet.. msync is the best we can do */
	if(NJ_ISSET(__nj_flags, NJ_PERSIST_HEAP))
	{
		/* If we have a persistant heap, store some stats info at the end */
		*((struct mem_stats *)&__nj_heap_tbl.table[__nj_heap_tbl.n_entries]) = __nj_mem_stats;
		msync((void *)__nj_heap_tbl.table, __nj_heap_tbl.n_entries*sizeof(struct heap_entry) + sizeof(struct mem_stats), 
				MS_SYNC);
	}

#ifdef _THREAD_SAFE
	if(NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
	{
		pthread_mutex_destroy(&__nj_pthread_lock);
		pthread_mutex_destroy(&__nj_heap_lock);
		pthread_mutex_destroy(&__nj_slab_lock);
		pthread_mutex_destroy(&__nj_init_lock);
		pthread_mutex_destroy(&__nj_list_lock);
	}
#endif

}

/**
 * Displays a string version of the sigaction si_code to the NJAMD error output.
 */
#ifdef HAVE_POSIX_SIGACTION
static void sigaction_perror(char *str, int sig, int si_code)
{
	switch(sig)
	{
		case SIGILL:	
# ifdef ILL_ILLOPC
			switch(si_code)
			{
				case ILL_ILLOPC: 
					__nj_eprintf("%sIllegal opcode\n", str); 
					break;
				case ILL_ILLOPN: 
					__nj_eprintf("%sIllegal operand\n", str); 
					break;
				case ILL_ILLADR: 
					__nj_eprintf("%sIllegal addressing mode\n", str); 
					break;
				case ILL_ILLTRP: 
					__nj_eprintf("%sIllegal trap\n", str); 
					break;
				case ILL_PRVOPC: 
					__nj_eprintf("%sPrivileged opcode\n", str); 
					break;
				case ILL_PRVREG: 
					__nj_eprintf("%sPrivileged register\n", str); 
					break;
				case ILL_COPROC: 
					__nj_eprintf("%sCopressesor error\n", str); 
					break;
				case ILL_BADSTK: 
					__nj_eprintf("%sInternal stack error\n", str); 
					break;
				default: 
					__nj_eprintf("%sUnknown code %d (sig %d). See sigaction(2)\n", str, si_code, sig);
			}
# endif
			break;
		case SIGFPE:
# ifdef FPE_INTDIV
			switch(si_code)
			{
				case FPE_INTDIV: 
					__nj_eprintf("%sInteger divide by zero\n", str); 
					break;
				case FPE_INTOVF: 
					__nj_eprintf("%sInteger overflow\n", str); 
					break;
				case FPE_FLTDIV: 
					__nj_eprintf("%sFloating point divide by zero\n", str); 
					break;
				case FPE_FLTOVF: 
					__nj_eprintf("%sFloating point overflow\n", str); 
					break;
				case FPE_FLTUND: 
					__nj_eprintf("%sFloating point underflow\n", str); 
					break;
				case FPE_FLTRES: 
					__nj_eprintf("%sFloating point inexact result\n", str); 
					break;
				case FPE_FLTINV: 
					__nj_eprintf("%sFloating point invalid operation\n", str); 
					break;
				case FPE_FLTSUB: 
					__nj_eprintf("%sSubscript out of range\n", str); 
					break;
				default: 
					__nj_eprintf("%sUnknown code %d (sig %d). See sigaction(2)\n", str, si_code, sig);
			}
# endif
			break;
		case SIGSEGV:
# ifdef SEGV_MAPERR
			switch(si_code)
			{
				case SEGV_MAPERR: 
					__nj_eprintf("%sAddress not mapped\n", str); 
					break;
				case SEGV_ACCERR: 
					__nj_eprintf("%sAccess to protected region\n", str); 
					break;
				default: 
					__nj_eprintf("%sUnknown code %d (sig %d). See sigaction(2)\n", str, si_code, sig);
			}
# endif
			break;

		case SIGBUS:
# ifdef BUS_ADRALN
			switch(si_code)
			{
				case BUS_ADRALN: 
					__nj_eprintf("%sInvalid address alignment\n", str); 
					break;
				case BUS_ADRERR: 
					__nj_eprintf("%sNon-existant physical address\n", str); 
					break;
				case BUS_OBJERR: 
					__nj_eprintf("%sObject specific hardware error\n", str); 
					break;
				default: 
					__nj_eprintf("%sUnknown code %d (sig %d). See sigaction(2)\n", str, si_code, sig); 
			}
# endif
			break;

		default: 
			__nj_eprintf("%sUnknown code %d (sig %d). See sigaction(2)\n", str, si_code, sig); 
			break;
	}
}
#endif /* POSIX_SIGACTION */

/**
 * Handles dispatching of various signals, deadly and nondeadly.
 *
 * NJAMD installs its own signal handlers for many signals. This function acts
 * as a relay point so that user's signal handlers still get called after the
 * necessary NJAMD action is taken. Its prototype is either that of a
 * sigaction handler or signal handler, depending on which is available.
 * 
 * @sideeffects Depending on the signal, either leaks are dumped (USR1), core
 * is dumped, and/or program termination
 */
#ifdef HAVE_POSIX_SIGACTION
void __nj_sig_dispatch(int sig, siginfo_t *info, void *arg)
#else
void  __nj_sig_dispatch(int sig)
#endif
{
	/** @FIXME: find something less arbitarty than 32 */
	u_long ret_addrs[32];
	

#if defined(DEBUG) || defined(EDEBUG)
	PRINTF("Last tracepoint reached in %s:%s at %d. Signaled at %p, with sig %d\n", 
			file, func, line, __builtin_return_address(0), sig);
	
#endif

	/* Print out messages */
	switch(sig)
	{
		case SIGSEGV:
			write(__nj_efd, "\nSegmentation fault (caught by NJAMD)\n", 38);
			break;
		case SIGBUS:
			write(__nj_efd, "\nBus Error (caught by NJAMD)\n", 29);
			break;
		case SIGFPE:
			write(__nj_efd, "\nFloating Point (caught by NJAMD)\n", 34);
			break;
		case SIGILL:
			write(__nj_efd, "\nIllegal Instruction (caught by NJAMD)\n", 39);
			break;
	}


	switch(sig)
	{
		case SIGUSR1:
			__nj_dump_leaks(TRACE_DEPTH);
			if(NJ_ISSET(__nj_flags, NJ_PERSIST_HEAP))
				msync((void *)__nj_heap_tbl.table, 
						__nj_heap_tbl.n_entries*sizeof(struct heap_entry),
					   	MS_SYNC);
			if(__nj_sig_hdlr[sig] != SIG_DFL)
				__nj_sig_hdlr[sig](sig);
#if HAVE_POSIX_SIGACTION
			else if(__nj_sig_actn[sig] != NULL)
				__nj_sig_actn[sig](sig, info, arg);
#endif
			break;
		/* Handle all the death signals */
		case SIGTERM:
		case SIGINT:
		case SIGQUIT:
		case SIGABRT:
			if(__nj_sig_hdlr[sig] != SIG_DFL)
				__nj_sig_hdlr[sig](sig);
#if HAVE_POSIX_SIGACTION
			else if(__nj_sig_actn[sig] != NULL)
				__nj_sig_actn[sig](sig, info, arg);
#endif
			else
			{
				__nj_cleanup();
				/* If we're aborting, we want a core file */
				if(sig != SIGABRT)
					_exit(sig);
			}
			break;
		case SIGILL:
		case SIGFPE:
		case SIGBUS:
		case SIGSEGV:
			/* FIXME: find someting less arbitrary than "32" */
			__nj_userspace_ret(ret_addrs, 32);
			__nj_print_stack_trace(ret_addrs, 32);
#if HAVE_POSIX_SIGACTION
			/* If the cause of the fault was not a raise(), print the real
			 * cause */
			if(info->si_code != SI_USER)
			{
				sigaction_perror("\nCause of fault: ", sig, info->si_code);
				__nj_eprintf("Address of fault is %p\n", info->si_addr);
			}
			/* TODO: if we ever allow the option of saving the state of all 
			 * allocs (CHK_FREE=infonly), then we will also wanna check the
			 * database if si_code = SEGV_MAPERR */
# ifdef SEGV_ACCERR
			if(info->si_code == SEGV_ACCERR)
# endif
				__nj_ptr_info((u_long)info->si_addr);
#endif

			if(NJ_ISSET(__nj_flags, NJ_DUMP_HARDCORE))
			{
				__nj_eprintf("NJAMD: Error, signal still handled in hardcore mode. mail mikepery@fscked.org\n");
			}

			__nj_cleanup();
			if(__nj_sig_hdlr[sig] != SIG_DFL)
				__nj_sig_hdlr[sig](sig);
#if HAVE_POSIX_SIGACTION
			else if(__nj_sig_actn[sig] != NULL)
				__nj_sig_actn[sig](sig, info, arg);
#endif
			else
			{
				if(NJ_ISSET(__nj_flags, NJ_DUMP_SOFTCORE))
					abort();
				else
					_exit(sig);
			}
			break;
		default:
			__nj_eprintf("NJAMD: Error, catching sig %d is not our job\n", sig);
			return;
	}

#ifndef HAVE_POSIX_SIGACTION
	signal(sig, __nj_sig_dispatch);
#endif
}

/**
 * Registers signal handlers.
 *
 * Register all the signals that would cause termination of the program, so we
 * can still spew some stats */
void __nj_register_handlers(void)
{
#ifdef HAVE_POSIX_SIGACTION
	struct sigaction act;

	memset(&act, 0, sizeof(act));
	act.sa_sigaction = __nj_sig_dispatch;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART | SA_SIGINFO;
#endif
	TR;
#ifdef HAVE_SANE_DLOPEN
# ifdef HAVE_POSIX_SIGACTION
	if(__njLibcSigaction)
	{
		__njLibcSigaction(SIGQUIT, &act, NULL);
		__njLibcSigaction(SIGABRT, &act, NULL);
		__njLibcSigaction(SIGINT, &act, NULL);
		__njLibcSigaction(SIGTERM, &act, NULL);
		__njLibcSigaction(SIGUSR1, &act, NULL);
		/* If hardcore dumping, don't handle signals */
		if(!NJ_ISSET(__nj_flags, NJ_DUMP_HARDCORE))
		{
			__njLibcSigaction(SIGILL, &act, NULL);
			__njLibcSigaction(SIGFPE, &act, NULL);
			__njLibcSigaction(SIGSEGV, &act, NULL);
			__njLibcSigaction(SIGBUS, &act, NULL);
		}
	}
# else
	if(__nj_LibcSignal)
	{
		__njLibcSignal(SIGQUIT, __nj_sig_dispatch);
		__njLibcSignal(SIGABRT, __nj_sig_dispatch);
		__njLibcSignal(SIGINT, __nj_sig_dispatch);
		__njLibcSignal(SIGTERM, __nj_sig_dispatch);
		__njLibcSignal(SIGUSR1, __nj_sig_dispatch);
		if(!NJ_ISSET(__nj_flags, NJ_DUMP_HARDCORE))
		{
			__njLibcSignal(SIGILL, __nj_sig_dispatch);
			__njLibcSignal(SIGFPE, __nj_sig_dispatch);
			__njLibcSignal(SIGBUS, __nj_sig_dispatch);
			__njLibcSignal(SIGSEGV, __nj_sig_dispatch);
		}
	}
# endif
#else
# ifdef HAVE_POSIX_SIGACTION
	sigaction(SIGABRT, &act, NULL);
	sigaction(SIGQUIT, &act, NULL);
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGTERM, &act, NULL);
	sigaction(SIGUSR1, &act, NULL);
	if(!NJ_ISSET(__nj_flags, NJ_DUMP_HARDCORE))
	{

		sigaction(SIGILL, &act, NULL);
		sigaction(SIGBUS, &act, NULL);
		sigaction(SIGFPE, &act, NULL);
		sigaction(SIGSEGV, &act, NULL);
	}
# else
	signal(SIGABRT, __nj_sig_dispatch);
	signal(SIGQUIT, __nj_sig_dispatch);
	signal(SIGINT, __nj_sig_dispatch);
	signal(SIGTERM, __nj_sig_dispatch);
	signal(SIGUSR1, __nj_sig_dispatch);
	if(!NJ_ISSET(__nj_flags, NJ_DUMP_HARDCORE))
	{
		signal(SIGILL, __nj_sig_dispatch);
		signal(SIGFPE, __nj_sig_dispatch);
		signal(SIGSEGV, __nj_sig_dispatch);
		signal(SIGBUS, __nj_sig_dispatch);
	}
# endif
#endif
	TR;
}

/** 
 * NJAMD's first round of intialization routines.
 * 
 * This is the first stage init code. It sets up the slabs, mutexes, handlers,
 * and status variables. There should be no races for this function, because
 * it is called before threads can be created by the user.
 *
 * @sideeffects The heap table is created here, along with all the mutexs. 
 */
void __nj_primary_init(void)
{
	char tmpf[] = "./njamd-zeroXXXXXX";
	const char *error;
	struct stat st;
	int i;

	if(NJ_ISSET(__nj_flags, NJ_RAN_PRIMARY))
		return;
	NJ_SET(__nj_flags, NJ_RAN_PRIMARY);
	

#ifndef NJ_PAGE_SIZE
	NJ_PAGE_SIZE = getpagesize();
	if((i = open("/dev/zero", O_RDONLY)) == -1)
	{
		__nj_perror("open");
		_exit(1);		
	}

	if((__nj_zero_page = mmap(NULL, NJ_PAGE_SIZE, PROT_READ, MAP_PRIVATE, i,
				0)) == -1)
	{
		__nj_perror("Mapping zero page");
		_exit(1);
	}
	close(i);
#endif
	
#ifndef NJ_PAGE_MASK
	NJ_PAGE_MASK = ~(NJ_PAGE_SIZE-1);
#endif
	
#ifndef NJ_PAGE_SHIFT
	for(i=0; i < sizeof(u_long)*8; i++)
	{
		if(NJ_PAGE_SIZE & (1<<i))
		{
			NJ_PAGE_SHIFT = i;
			break;
		}
	}
#endif

	TR;
	/* We do a little trick to save on dozens of ifdefs. mmaps that support
	 * anonymous mappings expect the fd to be -1. Those that don't expect no
	 * MAP_ANON flag and the fd to be /dev/zero for anonymous memory.
	 * Note: Some OS's (Linux especially) PREFAULT memory from /dev/zero,
	 * thats why we can't use it all the time.
	 */
	if(!MAP_ANON && (__nj_anonfd = open("/dev/zero", O_RDWR)) == -1)
	{
		__nj_perror("open");
		_exit(1);		
	}

	__nj_sbrk0 = (u_long)sbrk(0);

	__nj_efd = dup(STDERR_FILENO);
	TR;
	
	/* Workaround. If you have no permissions on your home directory and you
	 * call open this early, the program crashes mysteriously. Well, at least
	 * on glibc systems.. */
	stat("./", &st);

	i = geteuid();

	if(i != 0)
	{
		if(st.st_uid != i)
		{
			i = getegid();
			if(i != 0)
			{
				if(st.st_gid != i)
				{
					/* Directory world writable? */
					if(!(st.st_mode & S_IWOTH))
					{
						__nj_eprintf(__FUNCTION__ ": Can't create heap file: permission denied\n");
						_exit(1);
					}
				} /* Directory group writable? */
				else if(!(st.st_mode & S_IWGRP))
				{
					__nj_eprintf(__FUNCTION__ ": Can't create heap file: permission denied\n");
					_exit(1);
				}
			}
		} /* Directory user writable? */
		else if(!(st.st_mode & S_IWUSR))
		{
			__nj_eprintf(__FUNCTION__ ": Can't create heap file: permission denied\n");
			_exit(1);
		}
	}

	TR;
	/* Get a file descriptor for our zero page mapping */
#ifdef HAVE_WORKING_ERRNO
	if((__nj_zfd = mkstemp(tmpf)) == -1)
	{
		__nj_perror(__FUNCTION__": mkstemp");
		__nj_known_chk_free = SEGV_CHK_FREE;
	}
#else	
	if((__nj_zfd = open(tmpf, O_TRUNC|O_CREAT|O_RDWR, S_IRUSR|S_IWUSR)) == -1)
	{
		__nj_perror(__FUNCTION__ ": Can't create zero file");
		_exit(1);
	}
#endif

	write(__nj_zfd, __nj_zero_page, NJ_PAGE_SIZE);
	unlink(tmpf);
	TR;
	
	__nj_heap_tbl = __nj_new_heap_tbl();
	
	if(__nj_slab_alloc(0) == -1)
	{
		__nj_perror("NJAMD/"__FUNCTION__": slab alloc");
		nj_exit(1);
	}	

	TR;

	if((error = __njLibcFuncInit()) != NULL)
	{
		__nj_eprintf("NJAMD: dlopen: %s\n", error);
		nj_exit(1);
	}

	/* FBSD pthreads use malloc :( */
#ifdef _THREAD_SAFE
	pthread_mutex_init(&__nj_pthread_lock, NULL);
	pthread_mutex_init(&__nj_heap_lock, NULL);
	pthread_mutex_init(&__nj_slab_lock, NULL);
	pthread_mutex_init(&__nj_init_lock, NULL);
	pthread_mutex_init(&__nj_list_lock, NULL);
#endif
	TR;
}

/**
 * Second stage init called after main has been entered.
 *
 * The reason we try to wait for main is because on most platforms we don't
 * have any environment until main. This function also catalogs which address
 *
 * @sideeffects Many. Sets all the pref variables based on environment. Marks
 * 				the beginning and end of memory allocated by the preinit 
 * 				allocator.
 */
void __nj_secondary_init()
{
	/** Fuck this */
	struct rlimit rcore = {NJ_LIMIT_SOFTCORE, NJ_LIMIT_SOFTCORE};
	char *env_tmp;

#ifdef _THREAD_SAFE
	pthread_mutex_lock(&__nj_init_lock);
#endif

	/* FIXME: Wait for gdb's attachment if gdb < 4.18 */
	TR;	
	if(NJ_ISSET(__nj_flags, NJ_RAN_SECONDARY))
		return;

	/* record the region of memory for witch we used the preinit allocator.
	 * THe preinit allocator always uses slabs. */
	__nj_slab0 = (u_long)__nj_slab_tbl.tbl[0].addr;
	__nj_known_slab0 = (u_long)__nj_slab_tbl.tbl[0].next;

	__nj_mem_stats.alloc_type = DEFAULT_ALLOC_TYPE;

	if((env_tmp = getenv("NJAMD_PROT")) != NULL)
	{
		if(strncmp(env_tmp, "over",4) == 0)
		{
			__nj_known_alloc = __nj_overflow_alloc;
			__nj_known_free = __nj_overflow_free;
			__nj_mem_stats.alloc_type = ALLOC_OVER;
		}
		else if(strncmp(env_tmp, "under", 5) == 0)
		{
			__nj_known_alloc = __nj_underflow_alloc;
			__nj_known_free = __nj_underflow_free;
			__nj_mem_stats.alloc_type = ALLOC_UNDER;
		}
		else if(strncmp(env_tmp, "strict", 6) == 0)
		{
			__nj_known_alloc = __nj_sunderflow_alloc;
			__nj_known_free = __nj_sunderflow_free;
			__nj_mem_stats.alloc_type = ALLOC_SUNDER;
		}
		else if(strncmp(env_tmp, "none", 4) == 0)
		{
			__nj_prot_to_none();
			__nj_known_alloc = __nj_none_alloc;
			__nj_known_free = __nj_none_free;
			__nj_mem_stats.alloc_type = ALLOC_NONE;
		}
		else
			__nj_eprintf("NJAMD: Invalid malloc checking: %s\n", env_tmp);
	}
	
	TR;
	if((env_tmp = getenv("NJAMD_ALIGN")) != NULL)
	{
		TR;
		errno=0;
		__nj_align = strtol(env_tmp, NULL, 0);
		TR;
		if(errno)
		{
			TR;
			__nj_perror("NJAMD: invalid alignment");
			__nj_align = NJ_ALIGNMENT;
		}
	}

	if((env_tmp = getenv("NJAMD_DUMP_LEAKS_ON_EXIT")) != NULL)
	{
		TR;
		errno=0;
		__nj_display_depth = strtol(env_tmp, NULL, 0);
		TR;
		if(errno)
		{
			TR;
			__nj_perror("NJAMD: invalid depth");
			__nj_display_depth = TRACE_DEPTH;
		}

		if(__nj_display_depth > TRACE_DEPTH)
		{
			__nj_eprintf("NJAMD: depth %d > maximum %d\n", __nj_display_depth,
					TRACE_DEPTH);
			__nj_display_depth = TRACE_DEPTH;
		}
		/* Dumping stats comes free with dumping leaks */
		NJ_SET(__nj_flags, NJ_DUMP_STATS);
		
	}
	else if(getenv("NJAMD_DUMP_STATS_ON_EXIT") != NULL || getenv("NJAMD_DUMP_STATS") != NULL)
		NJ_SET(__nj_flags, NJ_DUMP_STATS);
	
	TR;
	if((env_tmp = getenv("NJAMD_DUMP_CORE")) != NULL)
	{
		if(strcmp(env_tmp, "soft") == 0)
		{
			NJ_SET(__nj_flags, NJ_DUMP_SOFTCORE);
		}
		else if(strcmp(env_tmp, "hard") == 0)
		{
			NJ_SET(__nj_flags, NJ_DUMP_HARDCORE);
		}
		else 
		{
			__nj_eprintf("NJAMD: Invalid porno^H^H^Hcoredump setting: %s\n", env_tmp);
		}
	}

	/* We have to set the limit on the core size if we're going
	   to map freed memory to no-access */
	if(!NJ_ISSET(__nj_flags, NJ_DUMP_HARDCORE))
		setrlimit(RLIMIT_CORE, &rcore);
	
	if((env_tmp = getenv("NJAMD_CHK_FREE")) != NULL)
	{
		if(strcmp(env_tmp, "error") == 0)
		{	
			__nj_known_chk_free = ERROR_CHK_FREE;
		}
		else if(strcmp(env_tmp, "none") == 0)
		{
			__nj_known_chk_free = NO_CHK_FREE;
		}
		else if(strcmp(env_tmp, "segv") == 0) 
		{
			__nj_known_chk_free = SEGV_CHK_FREE;
		}
		else if(strcmp(env_tmp, "nofree") == 0)
		{
			__nj_known_chk_free = NO_FREE;
		}
		else
		{
			__nj_eprintf("NJAMD: Invalid free checking: %s\n", env_tmp);
		}
	}
	else
	{
		__nj_known_chk_free = DEFAULT_CHK_FREE;
	}

	if((getenv("NJAMD_ALLOW_READ") != NULL))
		__nj_prot = PROT_READ;

	TR;

	/* Unlink the heap if the user didn't explicity say for it to persist */
	/* Stupid me an my lousy speeling. NJAMD < 0.7.1 spelled persistent
	 * wrong, so now we must maintain backward compatibility */
	if(!(getenv("NJAMD_PERSISTANT_HEAP") || getenv("NJAMD_PERSISTENT_HEAP")))
		unlink(__nj_heap_file);
	else
		NJ_SET(__nj_flags, NJ_PERSIST_HEAP);

	if(getenv("NJAMD_NO_TRACE"))
		NJ_SET(__nj_flags, NJ_NO_RETADDRS);

	if(getenv("NJAMD_ALLOW_FREE_0"))
		NJ_SET(__nj_flags, NJ_ALLOW_FREE0);

	if(getenv("NJAMD_ALLOW_MALLOC_0"))
		NJ_SET(__nj_flags, NJ_ALLOW_MALLOC0);

	if(getenv("NJAMD_TRACE_LIBS"))
		NJ_SET(__nj_flags, NJ_TRACE_LIBS);

	if((env_tmp = getenv("NJAMD_NO_FREE_INFO")))
		NJ_SET(__nj_flags, NJ_NO_FREE_INFO);
	
	/* We're runing the front end */
	if((env_tmp = getenv("NJAMD_FE_PORT")) != NULL)
	{
		int portno, sock;
		struct sockaddr_in dest;
		struct hostent *he;

		TR;	
		errno = 0;
		portno = strtol(env_tmp, NULL, 0);
		TR;
		if(!errno)
		{
			
			sock = socket(PF_INET, SOCK_STREAM, 0);
			dest.sin_family = AF_INET;
			dest.sin_port = htons(portno);
			memset(&(dest.sin_zero), 0, 8);
			
			/* Horrible, really, really horrible.. We have to do this because
			 * gethostbyname calls malloc on some platforms. This only happens
			 * once at the init of the program, so it's not worth installing a
			 * pthread_cond_wait for.
			 */
			NJ_SET(__nj_flags, NJ_RAN_SECONDARY);
			he = gethostbyname("localhost");
		
			dest.sin_addr = *((struct in_addr *) he->h_addr);
			TR;	
			if(connect(sock, &dest, sizeof(struct sockaddr_in)) == -1)
			{
				__nj_perror("Front end connect");	
			}
			else
			{
				close(__nj_efd);
				__nj_efd = sock;
				write(sock, __nj_itoa(getpid()), strlen(__nj_itoa(getpid())));
			}
		}
	}

#ifndef HAVE_ATTRIB_CONSTRUCT
	atexit(__nj_cleanup);
#endif

	/* MOVED FROM PRIMARY_INIT.
	 * Solaris has its head up its ass. It won't let you register signal
	 * handlers before main when using pthreads. Go figgure.
	 */
	__nj_register_handlers();

	/* This MUST be last, or other threads will think that secondary init has
	 * completed before it actually has. */
	NJ_SET(__nj_flags, NJ_RAN_SECONDARY);
	
#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_init_lock);
#endif

	TR;	
}

// vim:ts=4
