/** 
 * @file util.c 
 * Contains system utility fcns.
 *
 * Functions related to NJAMD's internal data structures and info gathering
 * cababilities lie here.
 *
 * Copyright (C) 2000 by Mike Perry.
 * Distributed WITHOUT WARRANTY under the GPL. See COPYING for details.
 * 
 */
#include <lib/njamd.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef HAVE_EXECINFO_H
# include <execinfo.h>
#endif

extern char **environ;

/**@{ @name Return address functions */
/* We must check the frame address as a hack... Sometimes high optimizations  
 * kill the frame pointer even if you don't specify -fomit-frame-pointer..
 * We wouldn't have to do this if the gcc people just followed their
 * documentation and returned NULL instead of segfaulting. I've reported this
 * bug and got no replies. *shrug*
 */
#if defined(USE_FRAME_ADDR_HACK)
# define RA(x)	case x: return (u_long)(__builtin_frame_address(x) ? __builtin_return_address(x) : 0)
#elif !defined(HAVE_DEADLY_RET_ADDR)
# define RA(x)	case x: return (u_long)__builtin_return_address(x)
#else /* __builtin_return_address is not only uselesss, its deadly */
# define RA(x)  case x: return NJ_RETADDR_NONE
#endif
/**
 * Gets a return address in the stack trace of the calling function.
 *
 * @param frame The number of function calls to traverse up the stack.
 * @returns A return address (aka calling address)
 */ 
static u_long return_address(int frame)
{
	TR;
	switch(frame+1)
	{
       RA(1);RA(2);RA(3);RA(4);RA(5);RA(6);RA(7);RA(8);RA(9);
	   RA(10);RA(11);RA(12);RA(13);RA(14);RA(15);RA(16);RA(17);RA(18);RA(19);
	   RA(20);RA(21);RA(22);RA(23);RA(24);RA(25);RA(26);RA(27);RA(28);RA(29);
	   RA(30);RA(31);RA(32);RA(33);RA(34);RA(35);RA(36);RA(37);RA(38);RA(39);
	   RA(40);RA(41);RA(42);RA(43);RA(44);RA(45);RA(46);RA(47);RA(48);RA(49);
	   RA(50);RA(51);RA(52);RA(53);RA(54);RA(55);RA(56);RA(57);RA(58);RA(59);
	   RA(60);RA(61);RA(62);RA(63);RA(64);RA(65);RA(66);RA(67);RA(68);RA(69);
	   RA(70);RA(71);RA(72);RA(73);RA(74);RA(75);RA(76);RA(77);RA(78);RA(79);
	   RA(80);RA(81);RA(82);RA(83);RA(84);RA(85);RA(86);RA(87);RA(88);RA(89);
	   RA(90);RA(91);RA(92);RA(93);RA(94);RA(95);RA(96);RA(97);RA(98);RA(99);
	   default:
	       return 0;
	}
	
}

/** 
 * Checks all the slab tables if the contain addr.
 * 
 * @param addr The address in question (usually passed to free)
 * @returns True if addr came from malloc
 */
int __nj_address_from_malloc(void *addr)
{
	int i;
	
	/* We can only perform this check if protection is NOT none. Cause
	 * in that case, we start to go outside the slabs */
	if(__nj_mem_stats.alloc_type == ALLOC_NONE)
		return 1;
	
	for(i=0; __nj_slab_tbl.tbl[i].addr != NULL; i++)
	{
		if((u_long)__nj_slab_tbl.tbl[i].addr < (u_long)addr 
				&& (u_long)addr < (u_long)__nj_slab_tbl.tbl[i].next)
			return 1;
	}
	return 0;
	
}

/**
 * Checks the given return address to see if the next one would possibly run
 * us off the call stack. There are several tricks this function does. First
 * off, the heap starts right above statically linked code in ELF, so the 
 * first sbrk in a program can be used to find out which return addresses are
 * in static user code.
 *
 * Also, we use the stubs in public.c to find out when to stop
 * backtracing once we are in the destructors (ie main would no longer be our
 * refrence point).
 * 
 * @param cmp The address in question (usually passed to free)
 * @param lower_bound Set to #__nj_sbrk0 if we want to stop at main, 0
 * 					  otherwise (ie if we know we're in the destructors)
 * @returns True if the next return address is not likely to knock us off the
 * call stack.
 */
static int next_address_valid(u_long cmp, u_long lower_bound)
{
	if(!cmp)
		return FALSE;
#ifdef HAVE_SANE_DLOPEN
# ifdef _THREAD_SAFE
	if(NJ_ISSET(__nj_flags, NJ_ACTIVE_PTHREADS))
	{
		/* only continue recursing if we're outside of main AND we're not inside
		 * the NJAMD pthread launching fucntion (which is the end of the call
		 * stack for pthreads */
		if(cmp > lower_bound
				&& !(((u_long)__nj_pthread_launch < cmp && cmp < (u_long)__nj_pthread_launch_end)
					|| ((u_long)__nj_pthread_exit_start < cmp && cmp < (u_long)__nj_pthread_exit_end)
					|| ((u_long)__nj_exit_start < cmp && cmp < (u_long)__nj_exit_end)))
			return TRUE;
	}
	else
# endif
	{
		/* Only continue recursing if we're outside of main and we're not in 
		 * exit */
		if(cmp > lower_bound
				&& !((u_long)__nj_exit_start < cmp && cmp < (u_long)__nj_exit_end))
			return TRUE;
	}
#else
	if(cmp > lower_bound)
		return TRUE;
#endif

	return FALSE;
}

/**
 * Gets a calltrace of specified length.
 * 
 * The actual addresses saved depend on a few flags, as well as the state of
 * the system, and the addresses next_address_valid() deems are terminal. If the
 * user doesn't want to trace libs (default), next_address_valid() is called
 * with #__nj_sbrk0 as the boundary, until it says that the return address is
 * below sbrk0. Otherwise, all addresses are stored until either we fill rets,
 * or next_address_valid() finds something else wrong with them.
 * 
 * @param rets The array of u_longs to fill with return addresses
 * @param depth The length of the array, or how far to trace.
 */
void __nj_userspace_ret(u_long *rets, int depth)
{
	int i, k;


	/* Three cases for not storing return address must be kept track of, so we
	 * know what to say/do when not displaying them */
	if(!NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN)) 
	{
		rets[0] = 0;
		return;
	}

	/* This flag isn't used in the library, but it is used in njamdpm */
	if(NJ_ISSET(__nj_flags, NJ_NO_RETADDRS))
	{
		rets[0] = NJ_RETADDR_NONE;
		return;
	}

	TR;

	if(NJ_ISSET(__nj_flags, NJ_ENTERED_DESTRUCT))
	{
		rets[0] = NJ_DESTRUCTOR;
		return;
	}

	rets[0] = (u_long)return_address(1);

	TR;
	
	if(NJ_ISSET(__nj_flags, NJ_TRACE_LIBS))
	{
		/* Only record return addresses where the next one is valid (as in
		 * from library space) */
		for(i = 1; i < depth; i++)
		{
#ifdef HAVE_WORKING_RET_ADDR
			if(next_address_valid(rets[i-1], 0))
#else
			if(next_address_valid(rets[i-1], __nj_sbrk0))
#endif
			{
				rets[i] = (u_long)return_address(i+1);
			}
			else
			{
				for(; i < depth; i++)
					rets[i] = 0;
				break;
			}
		}

	}
	else
	{
		/* Find the first invalid address and trace up from there */
		for(k=1; next_address_valid(rets[0], __nj_sbrk0); k++)
			rets[0] = (u_long)return_address(k+1);

#ifdef HAVE_WORKING_RET_ADDR
		/* fill in all return addresses with valid ones from usercode */
		for(i = 1; i < depth; i++)
		{
			if(next_address_valid(rets[i-1], 0))
			{
				rets[i] = (u_long)return_address(k+1);
				k++;
			}
			else
			{
				for(; i < depth; i++)
					rets[i] = 0;
				break;
			}
		}
#else
		for(i=1; i < depth; i++)
			rets[i] = 0;
#endif
	}

	TR;
	return;	
	
}

/** 
 * Attempts to detect if our execution path has flowed through
 * main yet. Again, we use the sbrk0 trick to determine if we have entered
 * statically compiled code. The problem is we don't know when to stop the
 * backtrace if we AREN'T past main.  Right now we stop when we find a
 * function that is below this library yet above the user segment, the
 * assumpion being that LD_PRELOADED and/or explicitly linked libs are the
 * lowest in the address space. This is about where __libc_start_main lies..
 * very hackish to say the least. ANy other ideas?
 *
 * This check must be done in the first place because most systems don't set
 * up the environment until just before main is called, so we need to check
 * for entrence into main before calling __nj_secondary_init()
 *
 * @returns True if we have crossed into statically linked code (ie most
 * likely have entered main, and have an environment)
 */
int __nj_entered_usercode()
{
#ifndef HAVE_DEADLY_RET_ADDR
	int i=1; /* Start at the return address of the caller of this function */
	u_long raddr;

	TR;
	
	for(;i<100;i++)
	{
		if((raddr = return_address(i)) < (u_long)__nj_primary_init && raddr > __nj_sbrk0)
			return FALSE;
		if(raddr < __nj_sbrk0)
			return TRUE;
	}
	TR;
	return FALSE;
#else
	return TRUE; /* __bultin_return_address is deadly. Punt and cross fingers */
#endif
}
/*@}*/

/**@{ @name Heap file functions */
/**
 * It's a new invention in computer science, the heap-stack :) but seriously 
 * folks.. 
 *
 * This monstrosity maintains the set of dirty heap entries that can be reused
 * when we reuse our address space. It's a stack because we're dealing with an
 * mmap'ed file, so temporal locality is a plus.
 *
 * @returns A previously used yet deallocated heap entry.
 * @sideeffects Modifies #__nj_heap_stack
 * @see __nj_heap_stack_push()
 */
struct heap_entry *__nj_heap_stack_pop(void)
{
	struct heap_entry *ret;

#ifdef _THREAD_SAFE
	pthread_mutex_lock(&__nj_heap_lock);
#endif

	if(__nj_heap_stack)
	{
		ret = __nj_heap_stack;
		__nj_heap_stack = __nj_heap_stack->u_addr.next;

#ifdef _THREAD_SAFE
		pthread_mutex_unlock(&__nj_heap_lock);
#endif
		return ret;
	}

#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_heap_lock);
#endif
	return NULL;
}

/**
 * The counterpart to __nj_heap_stack_pop(). This function is called when the
 * a memory segment is freed NO_FREE_INFO flag is set, to conserve the heap
 * table. 
 *
 * @param entry The heap entry of a freed segment.
 * @sideeffects Modifies __nj_heap_stack
 * @see __nj_heap_stack_pop()
 */

void __nj_heap_stack_push(struct heap_entry *entry)
{
#ifdef _THREAD_SAFE
	pthread_mutex_lock(&__nj_heap_lock);
#endif

	memset(entry, 0, sizeof(struct heap_entry));
	entry->u_addr.next = __nj_heap_stack;
	__nj_heap_stack = entry;

#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_heap_lock);
#endif
}

/**
 * Creates a new heap table in the event that one is full.
 * 
 * This does not need to be protected, because it's only called from inside 
 * the heap_lock mutex and the init mutex
 * 
 * @returns A new heap table to get heap entries from.
 * @see save_heap() __nj_new_heap_entry()
 * @sideeffects None.
 */
struct heap_tbl __nj_new_heap_tbl(void)
{
	struct heap_tbl new_table;
	int heapfd, i;
	
	/* We need a NEW heapfd for this.. we can't just expand the old one
	 * Even though errno should be valid by now, we want to keep the format
	 * consistant */
#ifdef HAVE_WORKING_ERRNO 
	snprintf(__nj_heap_file, sizeof(__nj_heap_file), "./njamd-%d-heap-%d", 
			getpid(), __nj_heap_top_dir.n_entries*HEAP_META_TBL_NENTRIES
			+ __nj_heap_mid_dir.n_entries+1);
#else
	/* if this is the first heap, we can't use snprintf cause errno
	 * doesn't work yet */
	if(__nj_heap_tbl.table == NULL)
		strncpy(__nj_heap_file, "./njamd-heap-0", sizeof(__nj_heap_file));
	else
		snprintf(__nj_heap_file, sizeof(__nj_heap_file), "./njamd-heap-%d", 
				__nj_heap_top_dir.n_entries*HEAP_META_TBL_NENTRIES 
			+ __nj_heap_mid_dir.n_entries+1);
#endif
	if((heapfd = open(__nj_heap_file, O_TRUNC|O_CREAT|O_RDWR, S_IRUSR|S_IWUSR)) == -1)
	{
		__nj_perror(__FUNCTION__ ": Can't create heap file");
		_exit(1);
	}

	/* What are we gonna do about initial memory state? */

	if((i = lseek(heapfd, HEAP_TBL_SIZE, SEEK_SET)) != HEAP_TBL_SIZE)
	{
		__nj_eprintf("Seeked %ld out of %ld\n", i, HEAP_TBL_SIZE);
		__nj_perror(__FUNCTION__": lseek");
		nj_exit(1);
	}

	/* Consider this an easter egg ;) */
	write(heapfd, "Mike Perry r0xx0rs", 1);

	lseek(heapfd, 0, SEEK_SET);

	if((new_table.table = mmap(NULL, HEAP_TBL_SIZE, PROT_READ|PROT_WRITE,
					MAP_SHARED, heapfd, 0)) == (void *)-1)
	{
		__nj_perror(__FUNCTION__": mmap");
		nj_exit(1);
	}

	close(heapfd);

	new_table.n_entries = 0;

	return new_table;
}

/**
 * Save the old heap into the meta table.
 *
 * Each heap table can only hold 32 bits worth of address space information.
 * When one fills, this function is called and stores it to a metatable called
 * __nj_heap_mid_dir.  Likewise, when the metatable fills, it is stored in a
 * supertable called #__nj_heap_top_dir. Note both metatables only store full
 * tables, so they are simply arrays of pointers to heap entries.
 *
 * @param tbl The current heap table to save.
 * @see __nj_new_heap_tbl() __nj_new_heap_entry()
 * @sideeffects Modifies the higher level heap metatables.
 */
static void save_heap(struct heap_tbl tbl)
{
	/* THis is not NENTRIES - 1, because we increment the count AFTER the
	 * fact.. so the last in use index is actually cnt - 1. Ie, cnt is an
	 * actual count */
	if(__nj_heap_mid_dir.table 
			&& __nj_heap_mid_dir.n_entries != HEAP_META_TBL_NENTRIES)
		goto out;

	if(__nj_heap_mid_dir.n_entries == HEAP_META_TBL_NENTRIES)
	{
		__nj_heap_mid_dir.n_entries = 0;

		if(!__nj_heap_top_dir.table)
		{	
			if((__nj_heap_top_dir.table = mmap(NULL, HEAP_META_TBL_SIZE, PROT_READ|PROT_WRITE,
							MAP_ANON, __nj_anonfd, 0)) == (void *)-1)
			{
				__nj_perror(__FUNCTION__": mmap");
				nj_exit(1);
			}
		}

		__nj_heap_top_dir.table[__nj_heap_top_dir.n_entries++] = __nj_heap_mid_dir.table;
	}
	

	if((__nj_heap_mid_dir.table = mmap(NULL, HEAP_META_TBL_SIZE, 
					PROT_READ|PROT_WRITE, MAP_ANON, __nj_anonfd, 
					0)) == (void *)-1)
	{
		__nj_perror(__FUNCTION__": mmap");
		nj_exit(1);
	}
	
out:

	/* notice the meta table is only a table of FULL heaps */
	__nj_heap_mid_dir.table[__nj_heap_mid_dir.n_entries++] = tbl.table;
}

/**
 * Creates a new heap entry describing a memory segment.
 *
 * @param start The start of the region, as returned from mmap.
 * @param len The length of the user's portion of the segment
 * @param retaddr A return address callstack recorded in malloc() and friends.
 * @returns A brand new heap entry filled with the given info.
 * @sideeffects May switch __nj_heap_tbl if it is full.
 *
 * @see save_heap() __nj_new_heap_tbl()
 */
struct heap_entry *__nj_new_heap_entry(void *start, int len, u_long *retaddr)
{
	struct heap_entry *ret;

	TR;
	/* If we're reusing the address space, then we have to use the list */
	if(!(ret = __nj_heap_stack_pop()))
	{
#ifdef _THREAD_SAFE
		pthread_mutex_lock(&__nj_heap_lock);
#endif

		if(__nj_heap_tbl.n_entries >= HEAP_TBL_NENTRIES)
		{
			/* save the heap state info at the top of the heap */
			*((struct mem_stats *)&__nj_heap_tbl.table[__nj_heap_tbl.n_entries]) = __nj_mem_stats;
			if(NJ_ISSET(__nj_flags, NJ_PERSIST_HEAP))
			{
				msync(__nj_heap_tbl.table, 
						__nj_heap_tbl.n_entries*sizeof(struct heap_entry) 
						    + sizeof(struct mem_stats), 
						MS_ASYNC);
			}
			/** @FIXME: this IS bad. check for other refs to heap_tbl,
			 * as well as fix the dumping of leaks at the end.
			 * I think this is mostly taken care of, except for updating
			 * njamdpm.c, which will be done by the util.c merge. */
			save_heap(__nj_heap_tbl);
			__nj_heap_tbl = __nj_new_heap_tbl();
			
		}
	
		ret = &__nj_heap_tbl.table[__nj_heap_tbl.n_entries++];
		memset(ret, 0, sizeof(struct heap_entry));
#ifdef _THREAD_SAFE
		pthread_mutex_unlock(&__nj_heap_lock);
#endif
	}

	TR;

	/** @FIXME: This code is duplicated in ALL the allocation functions */
	ret->u_addr.start = start;
	ret->ulen = len;
	ret->loc.freed[1] = NJ_RETADDR_NOTFREE;
	memcpy(ret->loc.alloced, retaddr, TRACE_DEPTH*sizeof(u_long));

	return ret;
	
}
/**
 * Checks to see if inf actually points to a heap table.
 *  
 * WHile this could be implemented very easily using for_all_heap_tables(), this
 * function is speed critical, so we do all the work inline.
 *  
 * @param inf The inf pointer to search the heap table(s) for.
 * @returns True if inf is in a table somewhere.
 * @see for_all_heap_tables()
 * @FIXME: this can become quite time consuming when an entire 64 bit address
 * 		   space is used.. Perhaps if there is a top_tbl, or even an m_tbl, we
 * 		   should just punt and return 1?
 */
int __nj_is_valid_heap_ptr(struct heap_entry *inf)
{
	register int i, j;

	/* the current heap isn't in the m tbl,	only full heaps go there */
	if(__nj_heap_tbl.table <= inf 
			&& inf < &__nj_heap_tbl.table[__nj_heap_tbl.n_entries])
		return 1;		
	
	/* don't worry, this will not deref a null ptr */
	for(i = 0; i < __nj_heap_mid_dir.n_entries; i++)
	{
		if(__nj_heap_mid_dir.table[i] <= inf 
				&& inf <= &__nj_heap_mid_dir.table[i][HEAP_TBL_NENTRIES-1])
			return 1;
	}
	
	/* if we're still here, try the other mmtables */
	for(i = 0; i < __nj_heap_top_dir.n_entries; i++)
	{
		for(j = 0; j < HEAP_META_TBL_NENTRIES; j++)
		{
			if(__nj_heap_top_dir.table[i][j] <= inf 
					&& inf <= &__nj_heap_top_dir.table[i][j][HEAP_TBL_NENTRIES-1])
				return 1;
		}       
	}

	/* if we're still here, we got nothing */
	return 0;
}

/**
 * A plugin function for for_all_heap_tables().
 *
 * This function takes the heap table give, and sees the addr passed in the
 * va_list is in that table. If So, return that heap entry.
 * 
 * @param tbl The heap table to search
 * @param va The va_list that contains the address to look for
 * @returns The heap entry that describes a segment containing the address
 * @see for_all_heap_tables() for_all_entries_in_tbl()
 */ 
static struct heap_entry *find_addr_in_tbl(struct heap_tbl *tbl, va_list ap)
{
	u_long addr = va_arg(ap, u_long);
	int i;
	
	for(i=0; i < tbl->n_entries; i++)
	{
		/* Make sure ulen is not zero */
		/* and if prot = none, then give the ulen, not the rounded amt */
		if(tbl->table[i].ulen && (u_long)tbl->table[i].u_addr.start <= addr 
				&& addr < (u_long)tbl->table[i].u_addr.start 
						  + ( (__nj_mem_stats.alloc_type == ALLOC_NONE) 
							  ? tbl->table[i].ulen 
							  : ROUND_TO_BLOCK(tbl->table[i].ulen)))
		{
			return &tbl->table[i];
		}
	}

	return NULL;
}

/**
 * Executes a function for each heap entry in a table.
 *
 * Continues to run until one of these functions returns non-null, at which
 * point execution ceases and that value is returned.
 *
 * @param tbl The table to use.
 * @param va A va_list that contains the function to execute and it's
 * 			 arguments
 * @returns A miscillanious pointer. Execution stops if non_null is returned.
 * @see for_all_heap_tables()
 */
static void *for_all_entries_in_tbl(void *tbl, va_list ap)
{
	nj_fcn_ptr_t do_fcn = va_arg(ap, nj_fcn_ptr_t);
	void *ret;
	int i;
	
	for(i = 0; i < ((struct heap_tbl *)tbl)->n_entries; i++)
	{
		if((ret = do_fcn(&((struct heap_tbl *)tbl)->table[i], ap)) != NULL)
			return ret;
	}

	return NULL;

}

/**
 * Iterates across all heap tables.
 *
 * Executes do_fcn on each table until some dern fool function returns
 * non-null.
 * @param do_fcn A function pointer to a function that takes a heap table and
 *               a va_list as arguments
 * @param ... Arguments to do_fcn
 * @returns The first non-null value from do_fcn.
 * @see __nj_dump_leaks()
 */
static void *for_all_heap_tables(nj_fcn_ptr_t do_fcn, ...)
{
	struct heap_tbl tbl;
	va_list ap;
	void *ret;
	int i, j;
	
	va_start(ap, do_fcn);

	if((ret = do_fcn(&__nj_heap_tbl, ap)) != NULL)
	{
		va_end(ap);
		return ret;
	}
	/* FIXME: if shit breaks, this is why: */
#ifdef NEED_VA_REWIND
	va_end(ap);
#endif

	/* All tables in the middle directory are full */
	tbl.n_entries = HEAP_META_TBL_NENTRIES;
	for(i = 0; i < __nj_heap_mid_dir.n_entries; i++)
	{
		tbl.table = __nj_heap_mid_dir.table[i];
#ifdef NEED_VA_REWIND
		va_start(ap, do_fcn);
#endif
		if((ret = do_fcn(&tbl, ap)) != NULL)
		{
			va_end(ap);
			return ret;
		}
#ifdef NEED_VA_REWIND
		va_end(ap);
#endif
	}

	for(i = 0; i < __nj_heap_top_dir.n_entries; i++)
	{
		/* All middle directories in the top directory are also full */
		for(j = 0; j < HEAP_META_TBL_NENTRIES; j++)
		{
			tbl.table = __nj_heap_top_dir.table[i][j];
#ifdef NEED_VA_REWIND
			va_start(ap, do_fcn);
#endif
			if((ret = do_fcn(&tbl, ap)) != NULL)
			{
				va_end(ap);
				return ret;
			}
#ifdef NEED_VA_REWIND
			va_end(ap);
#endif
		}
	}

	va_end(ap);
	return NULL;

}

/*@}*/

/**@{ @name Slab table functions */
/**
 * Allocates a new slab to cut memory blocks from.
 * 
 * Allocates slabs with a size backoff algorithm to compensate for
 * address space fragmentation.
 * 
 * @param idx The index in the slab table to put this slab at.
 * @returns -1 on failure, 0 otherwise. Errno will be set by the system call
 * that triggered it.
 * @sideffects __nj_slab_tbl.tbl has a new slab in it.
 * @see __nj_slab_fetch() __nj_slab_resize()
 */
int __nj_slab_alloc(int idx)
{
	int i;

	/* This function needn't be locked because it's only called on init
	 * (before threads exist) or inside another lock */

	/* This is a good time to write the heap table out to disk */
	if(NJ_ISSET(__nj_flags, NJ_PERSIST_HEAP))
		msync(__nj_heap_tbl.table, __nj_heap_tbl.n_entries*sizeof(struct heap_entry), 
				MS_ASYNC);

	TR;
	
	if(!__nj_slab_tbl.tbl)
	{
		TR;
		if((__nj_slab_tbl.tbl = mmap(NULL,
						SLAB_TABLE_SIZE,
						PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON,
						__nj_anonfd, 0)) == (void *)-1)
		{
			return -1;
		}
		memset(__nj_slab_tbl.tbl, 0, SLAB_TABLE_SIZE);
		__nj_slab_tbl.size = SLAB_TABLE_NENTRIES;
	}

	/* sucessively back off the requested size till we find something that
	 * fits or were asking for less than 2 pages. */
	for (i=0; (SLAB_SIZE >> i) >= (2 << NJ_PAGE_SHIFT); i++)
	{
		if((__nj_slab_tbl.tbl[idx].addr = mmap(NULL, SLAB_SIZE >> i, 
						PROT_READ|PROT_EXEC|PROT_WRITE, 
						MAP_PRIVATE|MAP_ANON, __nj_anonfd, 0)) != (void *)-1)
		{
			__nj_slab_tbl.tbl[idx].remainder = SLAB_SIZE >> i;
			__nj_slab_tbl.tbl[idx].next = __nj_slab_tbl.tbl[idx].addr;
	
			return 0;
		}
	}
	
	TR;
	return -1;	
}

/**
 * Resizes the slab table itself.
 *
 * The original slab table is large enough to hold 32bits of address space.
 * When it fills, double it and recopy it. The slab table is much smaller than
 * the heap table, so doubling it and recopying it is much more practical than
 * linear growth. Also, pointers point into the heap table, so we can't just
 * move it about willy nilly, but nothing points INTO the slab table. It only
 * describes the memory NJAMD uses.
 * 
 * @returns 0 on success, -1 and errno on failure.
 * @sideeffects __nj_slab_table is moved and doubled in size.
 * @see __nj_slab_fetch() __nj_slab_alloc()
 */
static int __nj_slab_resize(void)
{
	struct slab_tbl old_tbl = __nj_slab_tbl;

	TR;
	if(!__nj_slab_tbl.tbl)
	{
		TR;
		if((__nj_slab_tbl.tbl = mmap(NULL,
						SLAB_TABLE_SIZE,
						PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON,
						__nj_anonfd, 0)) == (void *)-1)
		{
			__nj_perror("NJAMD: Can't resize slab table");
			return -1;
		}
		memset(__nj_slab_tbl.tbl, 0, SLAB_TABLE_SIZE);
		__nj_slab_tbl.size = SLAB_TABLE_NENTRIES;
		return 0;
	}

	TR;
	/* double it */	
	__nj_slab_tbl.size <<= 1;
	
	if((__nj_slab_tbl.tbl = mmap(NULL,
					ROUND_TO_PAGE(__nj_slab_tbl.size*sizeof(struct slab_desc)),
					PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON,
					__nj_anonfd, 0)) == (void *)-1)
	{
		__nj_perror("NJAMD: Can't resize slab table");
		return -1;
	}
	memset(__nj_slab_tbl.tbl, 0,
			ROUND_TO_PAGE(__nj_slab_tbl.size*sizeof(struct slab_desc)));
	__nj_slab_tbl.size = ROUND_TO_PAGE(__nj_slab_tbl.size*sizeof(struct slab_desc))/sizeof(struct slab_desc);

	memcpy(__nj_slab_tbl.tbl, old_tbl.tbl, old_tbl.size*sizeof(struct slab_desc));
	
	munmap(old_tbl.tbl, old_tbl.size*sizeof(struct slab_desc));	

	return 0;
}

/**
 * Insert a chunk of freed memory into the free list.
 * Only used if NO_CHK_FREE is set, but this is a HUGE optimization. It allows
 * us to avoid munmapping, fetching off the slab, and reprotecting, and hoping
 * the OS is smart enough to defragment. Provies a 10X speedup.
 * 
 * @param segment beginning of the segment as returned from __nj_slab_fetch()
 * @param mem beginning of the writable portion of the segment
 * @param size the same size you would pass into __nj_slab_fetch() to get this 
 *        segment
 *
 * @see __nj_free_list_pop() __nj_slab_fetch()
 */
void __nj_free_list_push(void *segment, void *mem, size_t size) 
{ 
	int idx = ((size - NJ_PAGE_SIZE) >> NJ_PAGE_SHIFT)-1;

	if(idx > 3)
	{
		/* Ouch.. forgetting to munmap here would have been bad */
		if(munmap(segment, size) == -1)
		{
			__nj_perror(__FUNCTION__"/munmap");
			exit(1);
		}
		return;
	}

#ifdef _THREAD_SAFE
	pthread_mutex_lock(&__nj_list_lock);
#endif

	((struct free_list *)mem)->next = __nj_free_tbl[idx];
	__nj_free_tbl[idx] = mem;
	__nj_free_tbl[idx]->addr = segment;

#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_list_lock);
#endif

}

/**
 * Try to find a chunk of size size in the free list.
 * Only used if NO_CHK_FREE is set, but this is a HUGE optimization
 * 
 * @param size The size you would pass into __nj_slab_fetch()
 * @returns The beginning of the segment.
 * @see __nj_free_list_push()
 */
void *__nj_free_list_pop(size_t size)
{
	int idx = ((size - NJ_PAGE_SIZE) >> NJ_PAGE_SHIFT)-1;
	void *ret;

	/* We only cache alloced segments up to 3 pages */
	if(idx > 3)
		return NULL;

#ifdef _THREAD_SAFE
	pthread_mutex_lock(&__nj_list_lock);
#endif

	/* The segment itself IS the free list. So the linked list does not need
	 * to store the addr internally */
	if(__nj_free_tbl[idx])
	{
		ret = __nj_free_tbl[idx]->addr;
		__nj_free_tbl[idx] = __nj_free_tbl[idx]->next;
#ifdef _THREAD_SAFE
		pthread_mutex_unlock(&__nj_list_lock);
#endif
		return ret;
	}

#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_list_lock);
#endif
	return NULL;
	
}

/**
 * Fetch a section of memory from a slab or directly from a mmap.
 *
 * Having the memory mapped already in slabs is a large optimization that
 * saves us a system call on each alloc.
 * 
 * @param size The size to fetch (a multiple of NJ_PAGE_SIZE)
 * @returns The segment from the slab tables.
 * @sideeffects May fetch a new slab or resize the slab table
 * @see __nj_free_list_pop()
 */
void *__nj_slab_fetch(size_t size)
{
	int i;
	void *buf;

	TR;

	/* Stupid stupid stupid.. doing this here saves us NOTHING.. we want to
	 * save on the mprotect calls by using already mapped and protected
	 * segments */
#if 0
	if(__nj_known_chk_free == NO_CHK_FREE && 
		(buf = __nj_free_list_pop(size)) != NULL)
		return buf;
#endif

#ifdef _THREAD_SAFE
	pthread_mutex_lock(&__nj_slab_lock);
#endif

	for(i = 0; 
			i < __nj_slab_tbl.size 
			&& __nj_slab_tbl.tbl[i].addr != NULL 
			&& __nj_slab_tbl.tbl[i].remainder < size; 
			i++)
		;

	if(size > SLAB_SIZE)
	{
		for(; __nj_slab_tbl.tbl[i].addr != NULL; i++)
			;

		PRINTF("LOTS of memory requested (%d)\n", size);
			
		if((__nj_slab_tbl.tbl[i].addr = mmap(NULL, size, 
						PROT_READ|PROT_EXEC|PROT_WRITE, MAP_PRIVATE|MAP_ANON,
						__nj_anonfd, 0)) == (void *)-1)
			goto err;

		__nj_slab_tbl.tbl[i].remainder = 0;
		__nj_slab_tbl.tbl[i].next = __nj_slab_tbl.tbl[i].addr+size;

		buf = __nj_slab_tbl.tbl[i].addr;
#ifdef _THREAD_SAFE
		/* Kudos to Christian Gruner for catching this. */
		pthread_mutex_unlock(&__nj_slab_lock);
#endif
		
		return buf;		
	}

	TR;

	if(i >= __nj_slab_tbl.size)
	{
		TR;	
		if(__nj_slab_resize())
			goto err;
		
#ifdef _THREAD_SAFE
		pthread_mutex_unlock(&__nj_slab_lock);
#endif
		return __nj_slab_fetch(size);
	}

	TR;	
	if(__nj_slab_tbl.tbl[i].addr == NULL)
	{
		/* FIXME: For this block to be useful at all when NO_CHK_FREE is set, 
		 * we have to free the free list, because at this point, any free memory
		 * would be in there, yet in blocks to small to be useful.
		 *
		 * So the question is, can the MMU defragment the free list if we
		 * unmap the whole thing at this point?
		 * */
		if(__nj_slab_alloc(i) != 0)
				goto err;
	}
	
	__nj_slab_tbl.tbl[i].next += size;
	__nj_slab_tbl.tbl[i].remainder -= size;
	
	TR;		
	buf = __nj_slab_tbl.tbl[i].next - size;	
#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_slab_lock);
#endif
	return buf;

err:
	
#ifdef _THREAD_SAFE
	pthread_mutex_unlock(&__nj_slab_lock);
#endif
	
	if(errno == ENOMEM)
		__nj_out_of_memory();
	else
		__nj_perror("NJAMD: __nj_slab_fetch");
	nj_exit(errno);
	return NULL; /* Nope */
}
/*@}*/

/**@{ @name Information reporting routines */
/**
 * Prints out the stack trace. Resoloves symbols if possible.
 *
 * @param addr_array The call stack trace
 * @param depth The size of the trace
 */
void __nj_print_stack_trace(u_long *addr_array, int depth)
{
	int i;

	/* Change in NJAMDPM */
	if(NJ_ISSET(__nj_flags, NJ_NO_RETADDRS))
	{
		__nj_eprintf("\tCall stack not saved\n");
		return;
	}
	
	for(i=0; i < depth; i++)
	{
		if(addr_array[i] == 0)
			return;
		else if(addr_array[i] != NJ_DESTRUCTOR)
		{
#ifdef HAVE_BACKTRACE_SYMBOLS_FD
			__nj_eprintf("\tcalled from ");
			backtrace_symbols_fd( (void **)&addr_array[i], 1, __nj_efd );
#else
			__nj_eprintf("\tcalled from 0x%lx\n", addr_array[i]);
#endif
		}
		else
			__nj_eprintf("\tcalled from a destructor during program exit\n");
	}
	
}

/**
 * Prints information contained in a heap entry.
 * 
 * @param entry The entry to print
 * @param depth The depth of the call trace.
 * @returns 0 if this allocation should count towards the total, 0 otherwise
 * 			(ie a system allocation before main)
 */
int __nj_dump_entry(struct heap_entry *entry, int depth)
{

	if(!entry->ulen)
	{
		__nj_eprintf("\n0x%lx-0x%lx: Freed Memory\n", (u_long)entry->u_addr.start,
				(u_long)entry->u_addr.start + (__nj_mem_stats.alloc_type == ALLOC_NONE ? entry->ulen : ROUND_TO_BLOCK(entry->ulen)));
	}

	/* If zero, this allocation came before we entered main */
	if(!entry->loc.alloced[0])
	{
		TR;
		return 0;
	}
			
	__nj_eprintf("\n0x%lx-0x%lx: Aligned len %d\n   Allocation callstack:\n", 
			(u_long)entry->u_addr.start, 
			(u_long)entry->u_addr.start + (__nj_mem_stats.alloc_type == ALLOC_NONE ? entry->ulen : ROUND_TO_BLOCK(entry->ulen)), 
			entry->ulen);
	__nj_print_stack_trace(entry->loc.alloced, depth);
	
	if(!entry_is_freed(entry))
		__nj_eprintf("   Not Freed\n");
	else
	{
		__nj_eprintf("   Freed callstack:\n");
		__nj_print_stack_trace(entry->loc.freed, depth);
	}

	return 1;
}

/** @FIXME: None of this new code has worried about race conditions */

/** 
 * Provides information on a malloced pointer.
 * 
 * External symbol that can be called from gdb even if you don't have
 * PERSISTANT_HEAP set 
 * 
 * @param ptr The pointer to look up and display information about.
 */
void __nj_ptr_info(u_long ptr)
{
	struct heap_entry *entry;
	
	if((entry = for_all_heap_tables((nj_fcn_ptr_t)find_addr_in_tbl, ptr)) == NULL)
	{
		__nj_eprintf("Address 0x%lx not found in heap\n", (u_long)ptr);
		return;
	}
	__nj_dump_entry(entry, TRACE_DEPTH);
}



/**
 * A plugin function for for_all_entries_in_tbl()
 * 
 * @param ent The heap entry
 * @param va A va_list containing the depth of leaks to display
 * @returns NULL, since we want to iterate over all entries.
 * @see for_all_heap_tables() for_all_entries_in_tbl()
 */
static void *dump_entry_if_leaked(void *ent, va_list ap)
{
	if( ((struct heap_entry *)ent)->ulen != 0 
			&& !entry_is_freed((struct heap_entry *)ent))
		__nj_dump_entry(ent, va_arg(ap, int));

	/* return null, so we don't stop */
	return NULL; 
}

/**
 * Dumps all outstanding leaks to standard error
 *
 * @param depth The number of return addresses to display.
 */
void __nj_dump_leaks(int depth)
{
	if(depth > TRACE_DEPTH)
	{
		__nj_eprintf("Error, depth %d too large. Compiled max is %d\n", depth, TRACE_DEPTH);
		depth = TRACE_DEPTH;
	}

	for_all_heap_tables(for_all_entries_in_tbl, dump_entry_if_leaked, depth);

}

/** 
 * Update NJAMD statistics.
 * This function updates the statistics when amt memory is alloced
 * (> 0) or freed (< 0)
 * @param user The amount of user mem to add to stats (can be negative for free)
 * @param faulted The amount of total faulted pages (also negative for free)
 * @param as The amount of address space used (only negative if free and
 * 			 NO_CHK_FREE)
 */
void __nj_update_stats(int user, int faulted, int as)
{
	/* No stats for the system stuff, please. Also, don't bother keeping stats
	 * if no one'll ever see them. */
	if((!NJ_ISSET(__nj_flags, NJ_DUMP_STATS)
			   	&& !NJ_ISSET(__nj_flags, NJ_PERSIST_HEAP))
			|| !NJ_ISSET(__nj_flags, NJ_ENTERED_MAIN)) 
			return;

#if 0
	if(NJ_ISSET(__nj_flags, NJ_ENTERED_DESTRUCT))
		__nj_eprintf("Warning: free called in destructor. Please mail mikepery@fscked.org\n");
#endif

	__nj_leaked_user += user;
	__nj_leaked_faulted += faulted;
	__nj_mem_stats.as_used += as;

	if(user > 0)
	{
		__nj_mem_stats.total_user += user;
		__nj_mem_stats.total_faulted += faulted;
		__nj_mem_stats.total_allocs++;
		__nj_mem_stats.leaked_allocs++;

		if(__nj_leaked_user > __nj_mem_stats.peak_user)
		{
			__nj_mem_stats.peak_user = __nj_leaked_user;
			__nj_mem_stats.peak_faulted = __nj_leaked_faulted;
		}
		
		if(__nj_leaked_faulted - __nj_leaked_user > __nj_mem_stats.peak_overhead)
			__nj_mem_stats.peak_overhead = __nj_leaked_faulted - __nj_leaked_user;
	}
	else
		__nj_mem_stats.leaked_allocs--;
		
}

static char *__nj_byte_to_string(char *buf, int len, int bytes)
{
	if(bytes > 1<<SHIFT_MEGABYTE)
		snprintf(buf, len, "%.3f mB", bytes/(1024*1024.0));
	else if(bytes > 1024)
		snprintf(buf, len, "%.3f kB", bytes/1024.0);
	else
		snprintf(buf, len, "%d bytes", bytes);

	return buf;
}

/**
 * Print usage statistics.
 * 
 * This is just a quick little stat function so users can get statistics on
 * NJAMD performance as well as their program. Hopefully eventually we will
 * have some REALLY neat memory profiling cababilites that make this look like
 * trash.
 */
void __nj_usage_stats(void)
{
	/* Kinda silly, but what the hell. eprintf is much more expensive than
	 * stack space allocation. */
	char buf1[80], buf2[80], buf3[80], buf4[80], buf5[80], buf6[80], buf7[80];

	if(!__nj_mem_stats.total_allocs)
	{
		__nj_eprintf("\nNJAMD Totals:\n\tmalloc() was never called!\n");
		return;
	}

	__nj_eprintf(
			"\nNJAMD totals:\n\n"
			"\tAllocation totals:\t\t%d total, %d leaked\n"
			"\tLeaked User Memory:\t\t%s\n"
			"\tPeak User Memory:\t\t%s\n"
			"\tNJAMD Overhead at peak:\t\t%s\n" 
			"\tPeak NJAMD Overhead:\t\t%s\n" 
			"\tAverage NJAMD Overhead:\t\t%s per alloc\n" 
			"\tAddress space used:\t\t%s\n"
			"\tNJAMD Overhead at exit:\t\t%s\n\n",
			__nj_mem_stats.total_allocs, __nj_mem_stats.leaked_allocs, 
			__nj_byte_to_string(buf1, sizeof(buf1), __nj_leaked_user), 
			__nj_byte_to_string(buf2, sizeof(buf2), __nj_mem_stats.peak_user), 
			__nj_byte_to_string(buf3, sizeof(buf3), __nj_mem_stats.peak_faulted-__nj_mem_stats.peak_user),
			__nj_byte_to_string(buf4, sizeof(buf5), __nj_mem_stats.peak_overhead),
			__nj_byte_to_string(buf5, sizeof(buf5), (__nj_mem_stats.total_faulted-__nj_mem_stats.total_user)/__nj_mem_stats.total_allocs),
			__nj_byte_to_string(buf6, sizeof(buf6), __nj_mem_stats.as_used), 
			__nj_byte_to_string(buf7, sizeof(buf7), __nj_leaked_faulted-__nj_leaked_user)
				);
}

/**
 * A printf function guaranteed not to call malloc.
 *
 * This function writes out to __nj_efd, which is stderr in non-gui
 * operations. It doesn't call malloc, and the output size is limited to 1024
 * (local buffer)
 * @param format The string format
 * @param ... Arguments.
 */
void __nj_eprintf(const char *format, ...)
{
    char outbuf[1024];
    va_list ap;
    int len;
	va_start(ap, format);
	len = vsnprintf(outbuf, sizeof(outbuf), format, ap);
	if(len > 0)
		write(__nj_efd, outbuf, len);
	else
		write(__nj_efd, outbuf, strlen(outbuf));
	va_end(ap);
}
/*@}*/

/**
 * Cut losses on mapped memory if PROT=none is selected.
 *
 * This removes the mapped memory in the slab to save space if we're using
 * PROT_NONE.
 *
 * @siedeffects Unmaps unused slabs and slab table, as they won't be needed in
 * 			    PROT=none.
 */
void __nj_prot_to_none(void)
{
	/* Unmap the unused section of the slab */
    munmap((void *)__nj_known_slab0, SLAB_SIZE - (__nj_known_slab0 - __nj_slab0));
	/* Unmap the slab table. We won't be using it no mo */
	munmap((void *)__nj_slab_tbl.tbl,
		            ROUND_TO_PAGE(__nj_slab_tbl.size*sizeof(struct slab_desc)));
}

#ifdef _THREAD_SAFE
/**
 * Launch a new thread. Using this function we can detect the end of a call
 * stack trace, so we don't fly off the improperly terminated stacks of some
 * pthread implementations.
 *
 * Isn't this fantastic? I'm pretty proud of this little hack actually :)
 *
 * @param arg A "wrapper" argument contianing the function to call as well as
 * 		      it's argument.
 * @returns Whatever the actual pthread function returns.
 * @sideeffects Releases the __nj_pthread_lock created in pthread_create() to
 * 			    protect it's argument.
 * @see pthread_create() next_address_valid() __nj_userspace_ret()
 */
void *__nj_pthread_launch(void *arg)
{
	struct pthread_arg_wrapper warg = *(struct pthread_arg_wrapper *)arg;

	/* This lock initiated in our wrapped version of pthread_create, to
	 * protect the static variable passed to us */
	pthread_mutex_unlock(&__nj_pthread_lock);
	TR;
	PRINTF("warg: %p, start: %p, real arg: %p\n", arg, warg.start,
			warg.real_arg);
	return warg.start(warg.real_arg);
}

/** Placeholder to detrmine the size of __nj_pthread_launch */
void __nj_pthread_launch_end(void)
{
	return;
}
#endif

/** 
 * Prints an error message then exits if we run out of memory. 
 */
void __nj_out_of_memory(void)
{
#ifdef __linux__
	int fd;
	char did_oc = 0;
	char mmc_str[32];
	int mmc = 0;
		
	if((fd = open(LINUX_PROC_MAX_MAP_COUNT, O_RDONLY) != -1))
	{
		read(fd, mmc_str, sizeof(mmc_str));
		mmc = strtol(mmc_str, NULL, 0);
		close(fd);
	}
	
	if((fd = open(LINUX_PROC_OVERCOMMIT, O_RDONLY)) != -1)
	{
		read(fd, &did_oc, 1);
		did_oc -= '0';
		close(fd);
	}

	switch(__nj_known_chk_free)
	{
		case ERROR_CHK_FREE:
			if(did_oc && mmc > 65536)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nSet NJAMD_CHK_FREE=segv or none\n");
			else if(did_oc)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nDid you try the proc_map patch?\n");
			else
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nRun sysctl -w vm.overcommit_memory=1 as root, or set NJAMD_CHK_FREE=segv or none\n");
			break;
		case SEGV_CHK_FREE:
			if(did_oc && mmc > 65536)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nSet NJAMD_CHK_FREE=none\n");
			else if(did_oc)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nDid you try the proc_map patch?\n");
			else
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nRun sysctl -w vm.overcommit_memory=1 as root, or set NJAMD_CHK_FREE=segv or none\n");
			break;
		default:
			if(did_oc && mmc > 65536)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\n");
			else if(did_oc)
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nDid you try the proc_map patch?\n");
			else
				__nj_eprintf("\nNJAMD: Address space exhaustion.\nRun sysctl -w vm.overcommit_memory=1 and try the proc_map patch.\n");
			break;
	}
#else
	switch(__nj_known_chk_free)
	{
		case ERROR_CHK_FREE:
			__nj_eprintf("\nNJAMD: Address space exhaustion.\nSet NJAMD_CHK_FREE=segv or none\n");
			break;
		case SEGV_CHK_FREE:
			__nj_eprintf("\nNJAMD: Address space exhaustion.\nSet NJAMD_CHK_FREE=none\n");
			break;
		default:
			__nj_eprintf("\nNJAMD: Address space exhaustion.\nDid you try telling your OS to overcommit virtual memory?\n");
			break;
	}

#endif
	if(NJ_ISSET(__nj_flags, NJ_DUMP_STATS))
		__nj_usage_stats();
	/* Turn off printing of usage stats here */
	nj_exit(ENOMEM);

}

/**@{ @name Thread UNSAFE
 * @FIXME: the following functions are NOT threadsafe.. They are just error
 * reporting functions though. There really is no danger, in that they are
 * only called when something is ALREADY fucked up 
 */

/**
 * Converts integer to string
 * @param i Int to convert
 * @returns Statically allocated character pointer
 */
char *__nj_itoa(int i)
{
	static char a[16];
	snprintf(a, 15, "%d", i);
	return a;
}

/**
 * Converts pointer to string
 * @param i pointer to convert
 * @returns Statically allocated character pointer
 */
char *__nj_ptoa(void *i)
{
	static char a[16];
	snprintf(a, 15, "0x%lx", (u_long)i);
	return a;
}

/**
 * Internal perror that will never call malloc.
 * 
 * We needed our own printf and perror that write immeditaly, and can be
 * guaranteed not to use malloc ever.
 *
 * @param s Error prefix string 
 */
void __nj_perror(const char *s)
{
	__nj_eprintf("%s: %s\n", s, strerror(errno));
}
/*@}*/
// vim:ts=4
